from pygbif.gbifutils import check_data, len2, gbif_baseurl, gbif_GET


def name_usage(
    key=None,
    name=None,
    data="all",
    language=None,
    datasetKey=None,
    uuid=None,
    sourceId=None,
    rank=None,
    shortname=None,
    limit=100,
    offset=None,
    **kwargs
):
    """
    Lookup details for specific names in all taxonomies in GBIF.

    :param key: [fixnum] A GBIF key for a taxon
    :param name: [str] Filters by a case insensitive, canonical namestring,
         e.g. 'Puma concolor'
    :param data: [str] The type of data to get. Default: ``all``. Options: ``all``,
        ``verbatim``, ``name``, ``parents``, ``children``,
        ``related``, ``synonyms``, ``descriptions``, ``distributions``, ``media``,
        ``references``, ``speciesProfiles``, ``vernacularNames``, ``typeSpecimens``,
        ``root``
    :param language: [str] Language. Expects a ISO 639-1 language codes using 2 lower
        case letters. Languages returned are 3 letter codes. The language parameter
        only applies to the ``/species``, ``/species/{int}``,
        ``/species/{int}/parents``, ``/species/{int}/children``, ``/species/{int}/related``,
        ``/species/{int}/synonyms`` routes (here routes are determined by the ``data``
        parameter).
    :param datasetKey: [str] Filters by the dataset's key (a uuid)
    :param uuid: [str] A uuid for a dataset. Should give exact same results as datasetKey.
    :param sourceId: [fixnum] Filters by the source identifier.
    :param rank: [str] Taxonomic rank. Filters by taxonomic rank as one of:
            ``CLASS``, ``CULTIVAR``, ``CULTIVAR_GROUP``, ``DOMAIN``, ``FAMILY``, ``FORM``, ``GENUS``, ``INFORMAL``,
            ``INFRAGENERIC_NAME``, ``INFRAORDER``, ``INFRASPECIFIC_NAME``, ``INFRASUBSPECIFIC_NAME``,
            ``KINGDOM``, ``ORDER``, ``PHYLUM``, ``SECTION``, ``SERIES``, ``SPECIES``, ``STRAIN``, ``SUBCLASS``, ``SUBFAMILY``,
            ``SUBFORM``, ``SUBGENUS``, ``SUBKINGDOM``, ``SUBORDER``, ``SUBPHYLUM``, ``SUBSECTION``, ``SUBSERIES``,
            ``SUBSPECIES``, ``SUBTRIBE``, ``SUBVARIETY``, ``SUPERCLASS``, ``SUPERFAMILY``, ``SUPERORDER``,
            ``SUPERPHYLUM``, ``SUPRAGENERIC_NAME``, ``TRIBE``, ``UNRANKED``, ``VARIETY``
    :param shortname: [str] A short name..need more info on this?
    :param limit: [fixnum] Number of records to return. Default: ``100``. Maximum: ``1000``. (optional)
    :param offset: [fixnum] Record number to start at. (optional)

    References: See http://www.gbif.org/developer/species#nameUsages for details

    Usage::

            from pygbif import species

            species.name_usage(key=1)

            # Name usage for a taxonomic name
            species.name_usage(name='Puma', rank="GENUS")

            # All name usages
            species.name_usage()

            # References for a name usage
            species.name_usage(key=2435099, data='references')

            # Species profiles, descriptions
            species.name_usage(key=5231190, data='speciesProfiles')
            species.name_usage(key=5231190, data='descriptions')
            species.name_usage(key=2435099, data='children')

            # Vernacular names for a name usage
            species.name_usage(key=5231190, data='vernacularNames')

            # Limit number of results returned
            species.name_usage(key=5231190, data='vernacularNames', limit=3)

            # Search for names by dataset with datasetKey parameter
            species.name_usage(datasetKey="d7dddbf4-2cf0-4f39-9b2a-bb099caae36c")
    """
    args = {
        "language": language,
        "name": name,
        "datasetKey": datasetKey,
        "rank": rank,
        "sourceId": sourceId,
        "limit": limit,
        "offset": offset,
    }
    data_choices = [
        "all",
        "verbatim",
        "name",
        "parents",
        "children",
        "related",
        "synonyms",
        "descriptions",
        "distributions",
        "media",
        "references",
        "speciesProfiles",
        "vernacularNames",
        "typeSpecimens",
        "root",
    ]
    check_data(data, data_choices)
    if len2(data) == 1:
        return name_usage_fetch(data, key, shortname, uuid, args, **kwargs)
    else:
        return [name_usage_fetch(x, key, shortname, uuid, args, **kwargs) for x in data]


def name_usage_fetch(x, key, shortname, uuid, args, **kwargs):
    if x != "all" and key is None:
        raise TypeError("You must specify `key` if `data` does not equal `all`")

    if x == "all" and key is None:
        url = gbif_baseurl + "species"
    else:
        if x == "all" and key is not None:
            url = gbif_baseurl + "species/" + str(key)
        else:
            if x in [
                "verbatim",
                "name",
                "parents",
                "children",
                "related",
                "synonyms",
                "descriptions",
                "distributions",
                "media",
                "references",
                "speciesProfiles",
                "vernacularNames",
                "typeSpecimens",
            ]:
                url = gbif_baseurl + "species/%s/%s" % (str(key), x)
            else:
                if x == "root":
                    url = gbif_baseurl + "species/%s/%s" % (uuid, shortname)

    res = gbif_GET(url, args, **kwargs)
    return res



"""Tests for species module - name_usage methods"""
import vcr
from pygbif import species


@vcr.use_cassette("test/vcr_cassettes/test_name_backbone.yaml")
def test_name_backbone():
    "species.name_backbone - basic test"
    res = species.name_backbone(scientificName="Calopteryx splendens")
    assert dict == res.__class__
    assert 5 == len(res)
    assert "Calopteryx splendens (Harris, 1780)" == res["usage"]["name"]
    assert list(res.keys()) == ['usage', 'classification', 'diagnostics', 'additionalStatus', 'synonym']

@vcr.use_cassette("test/vcr_cassettes/test_name_backbone_verbose.yaml")
def test_name_backbone_verbose():
    "species.name_backbone - verbose test"
    res = species.name_backbone(scientificName="Calopteryx", verbose=True)
    assert dict == res.__class__
    assert list(res.keys()) == ['diagnostics', 'synonym']
    assert list(res["diagnostics"]) == ['matchType', 'issues', 'confidence', 'note', 'timeTaken', 'alternatives', 'timings']
    assert len(res["diagnostics"]["alternatives"]) > 5

@vcr.use_cassette("test/vcr_cassettes/test_name_backbone_class.yaml")
def test_name_backbone_class():
    "species.name_backbone - class test"
    res = species.name_backbone(class_="Insecta")
    assert dict == res.__class__
    assert 4 == len(res)
    assert "Insecta" == res["usage"]["name"]
    assert list(res.keys()) == ['usage', 'classification', 'diagnostics', 'synonym']

@vcr.use_cassette("test/vcr_cassettes/test_name_backbone_checklistKey.yaml")
def test_name_backbone_checklistKey():
    "species.name_backbone - checklistKey test"
    res = species.name_backbone(scientificName="Calopteryx splendens", checklistKey="7ddf754f-d193-4cc9-b351-99906754a03b")
    assert dict == res.__class__
    assert res["usage"]["key"] == "Q2M4"
    assert 5 == len(res)
    assert "Calopteryx splendens" == res["usage"]["canonicalName"]
    assert list(res.keys()) == ['usage', 'classification', 'diagnostics', 'additionalStatus', 'synonym']




import requests
import re
import pygbif

# import requests_cache
# from requests_cache.core import remove_expired_responses
# import os.path
# import tempfile

# CACHE_FILE = os.path.join(tempfile.gettempdir(), 'pygbif_requests_cache')
# expire = 300
# backend = "sqlite"
# requests_cache.install_cache(cache_name=CACHE_FILE, backend=backend, expire_after=expire)
# remove_expired_responses()


class NoResultException(Exception):
    pass


def gbif_search_GET(url, args, **kwargs):
    # if args['geometry'] != None:
    #   if args['geometry'].__class__ == list:
    #     b = args['geometry']
    #     args['geometry'] = geometry.box(b[0], b[1], b[2], b[3]).wkt
    out = requests.get(url, params=args, **kwargs)
    out.raise_for_status()
    stopifnot(out.headers["content-type"])
    return out.json()


def gbif_GET(url, args, **kwargs):
    out = requests.get(url, params=args, headers=make_ua(), **kwargs)
    out.raise_for_status()
    stopifnot(out.headers["content-type"])
    return out.json()

def gbif_GET_raw(url):
    out = requests.get(url)
    return out.content

def gbif_GET_map(url, args, ctype, **kwargs):
    out = requests.get(url, params=args, headers=make_ua(), **kwargs)
    out.raise_for_status()
    stopifnot(out.headers["content-type"], ctype)
    return out


def gbif_GET_write(url, path, **kwargs):
    out = requests.get(url, headers=make_ua(), stream=True, **kwargs)
    out.raise_for_status()
    if out.status_code == 200:
        with open(path, "wb") as f:
            for chunk in out.iter_content(chunk_size=1024):
                if chunk:
                    f.write(chunk)
    # FIXME: removing response content-type check for now, maybe add later
    # ctype = "application/octet-stream"
    # if not re.match(ctype, out.headers["content-type"]):
    #     raise NoResultException("content-type did not contain '%s'" % ctype)
    return path


def gbif_POST(url, body, **kwargs):
    head = make_ua()
    out = requests.post(url, json=body, headers=head, **kwargs)
    out.raise_for_status()
    stopifnot(out.headers["content-type"])
    return out.json()


def gbif_DELETE(url, body, **kwargs):
    head = make_ua()
    out = requests.delete(url, json=False, headers=head, **kwargs)
    out.raise_for_status()
    return out.status_code == 204


def stopifnot(x, ctype="application/json"):
    if x != ctype:
        raise NoResultException("content-type did not equal " + ctype)


def stop(x):
    raise ValueError(x)


def make_ua():
    return {
        "user-agent": "python-requests/"
        + requests.__version__
        + ",pygbif/"
        + pygbif.__version__
    }


def is_none(x):
    return x.__class__.__name__ == "NoneType"


def is_not_none(x):
    return x.__class__.__name__ != "NoneType"


gbif_baseurl = "https://api.gbif.org/v1/"

requests_argset = [
    "timeout",
    "cookies",
    "auth",
    "allow_redirects",
    "proxies",
    "verify",
    "stream",
    "cert",
]


def bn(x):
    if x:
        return x
    else:
        return None


def parse_results(x, y):
    if y.__class__.__name__ != "NoneType":
        if y.__class__ != dict:
            return x
        else:
            if "endOfRecords" in x.keys():
                return x["results"]
            else:
                return x
    else:
        return x["results"]


def check_data(x, y):
    if len2(x) == 1:
        testdata = [x]
    else:
        testdata = x

    for z in testdata:
        if z not in y:
            raise TypeError(z + " is not one of the choices")


def len2(x):
    if isinstance(x, int):
        return len([x])
    elif isinstance(x, str):
        return len([x])
    else:
        return len(x)


def stuff(**kwargs):
    return kwargs


def check_param_lens(**kwargs):
    tmp = {k: v for k, v in kwargs.items() if v is not None}
    for k, v in tmp.items():
        if len2(v) > 1:
            raise TypeError(k + " must be length 1")


def get_meta(x):
    if has_meta(x):
        return {z: x[z] for z in ["offset", "limit", "endOfRecords"]}
    else:
        return None


def has_meta(x):
    if x.__class__ != dict:
        return False
    else:
        tmp = [y in x.keys() for y in ["offset", "limit", "endOfRecords"]]
        return True in tmp


def has(str, pattern):
    w = re.search(pattern, str)
    return w is not None


def bool2str(x):
    if x is not None:
        z = str(x).lower()
        return z
    else:
        return x



"""Tests for occurrences module - get methods"""
import vcr
from pygbif import occurrences


@vcr.use_cassette("test/vcr_cassettes/test_get.yaml")
def test_get():
    "occurrences.get - basic test"
    res = occurrences.get(key=142316497)
    assert "dict" == res.__class__.__name__
    assert len(res) > 30
    assert 142316497 == res["key"]


@vcr.use_cassette("test/vcr_cassettes/test_get_verbatim.yaml")
def test_get_verbatim():
    "occurrences.get_verbatim - basic test"
    res = occurrences.get_verbatim(key=142316497)
    assert "dict" == res.__class__.__name__
    assert len(res) > 20
    assert 142316497 == res["key"]


@vcr.use_cassette("test/vcr_cassettes/test_get_fragment.yaml")
def test_get_fragment():
    "occurrences.get_fragment - basic test"
    res = occurrences.get_fragment(key=1986620884)
    assert "dict" == res.__class__.__name__
    assert "HumanObservation" == res["basisOfRecord"]



"""Tests for occurrences module - download_describe"""
from pygbif import occurrences as occ
import vcr

@vcr.use_cassette("test/vcr_cassettes/test_download_describe.yaml")
def test_download_describe():
    "occurrences.download_describe - basic usage"
    res=occ.download_describe("simpleCsv") 
    assert dict == res.__class__ 
    assert len(res["fields"]) >= 50 # unlikely to get smaller
    assert "gbifID" == res["fields"][0]["name"]

def test_download_describe_fails_well():
    "occurrences.download_describe - fail test"
    try:
        res=occ.download_describe("dog")
    except Exception as e:
        assert str(e) == "format not in list of acceptable formats"



from pygbif import occurrences as occ
import vcr

@vcr.use_cassette('test/vcr_cassettes/test-occurrences-download_citation.yaml')
def test_download_citation():
    res=occ.download_citation("0235283-220831081235567")
    assert "str" == res.__class__.__name__
    "GBIF.org (2 January 2023) GBIF Occurrence Download https://doi.org/10.15468/dl.29wbtx" == res 

def test_download_citation_failswell():
    try:
        occ.download_citation("dog")
    except ValueError as e:
        assert str(e) == "key must be a GBIF download key"




"""Tests for registry module - installations methods"""
import vcr

from pygbif import registry


@vcr.use_cassette("test/vcr_cassettes/test_installations.yaml")
def test_installations():
    "registry.installations - basic test"
    res = registry.installations()
    assert dict == res.__class__
    assert 2 == len(res)
    assert 100 == len(res["data"])
    assert ["data", "meta"] == sorted(res.keys())


@vcr.use_cassette("test/vcr_cassettes/test_installations_limit.yaml")
def test_installations_limit():
    "registry.installations - limit param "
    res = registry.installations(limit=5)
    assert dict == res.__class__
    assert 5 == len(res["data"])


@vcr.use_cassette("test/vcr_cassettes/test_installations_uuid.yaml")
def test_installations_uuid():
    "registry.installations - with a uuid"
    res = registry.installations(uuid="b77901f9-d9b0-47fa-94e0-dd96450aa2b4")
    assert dict == res.__class__
    assert 2 == len(res)
    assert 16 == len(res["data"])
    assert dict == res["data"]["endpoints"][0].__class__
    assert "identifiers" in res["data"].keys()



"""Tests for occurrences module - count methods"""
import pytest
import vcr
from pygbif import occurrences

brecord_res = [
    "HUMAN_OBSERVATION",
    "PRESERVED_SPECIMEN", 
    "MATERIAL_SAMPLE", 
    "OBSERVATION", 
    "MACHINE_OBSERVATION", 
    "OCCURRENCE", 
    "FOSSIL_SPECIMEN", 
    "MATERIAL_CITATION", 
    "LIVING_SPECIMEN"
]

year_res = [
    "1991",
    "1990",
    "1993",
    "1992",
    "1995",
    "1994",
    "1997",
    "1996",
    "1999",
    "1998",
    "2000",
]


@vcr.use_cassette("test/vcr_cassettes/test_count.yaml")
def test_count():
    "occurrences.count - basic test"
    res = occurrences.count(taxonKey=3329049)
    assert int == res.__class__


def test_count_param_length():
    "occurrences.count_param_length"
    with pytest.raises(TypeError):
        occurrences.count(datasetKey=["foo", "bar"])


@vcr.use_cassette("test/vcr_cassettes/test_count_basisofrecord.yaml")
def test_count_basisofrecord():
    "occurrences.count_basisofrecord - basic test"
    res = occurrences.count_basisofrecord()
    assert dict == res.__class__
    assert 9 == len(res)
    assert sorted(brecord_res) == sorted(res.keys())


@vcr.use_cassette("test/vcr_cassettes/test_count_year.yaml")
def test_count_year():
    "occurrences.count_year - basic test"
    res = occurrences.count_year(year="1990,2000")
    assert dict == res.__class__
    assert 11 == len(res)
    assert sorted(year_res) == sorted(res.keys())


@vcr.use_cassette("test/vcr_cassettes/test_count_datasets.yaml")
def test_count_datasets():
    "occurrences.count_datasets - basic test"
    res = occurrences.count_datasets(country="DE")
    assert dict == res.__class__
    assert str == str(list(res.keys())[0]).__class__


@vcr.use_cassette("test/vcr_cassettes/test_count_countries.yaml")
def test_count_countries():
    "occurrences.count_countries - basic test"
    res = occurrences.count_countries(publishingCountry="DE")
    assert dict == res.__class__
    assert str == str(list(res.keys())[0]).__class__
    assert int == list(res.values())[0].__class__


@vcr.use_cassette("test/vcr_cassettes/test_count_schema.yaml")
def test_count_schema():
    "occurrences.count_schema - basic test"
    res = occurrences.count_schema()
    assert list == res.__class__
    assert dict == res[0].__class__
    assert "dimensions" == list(res[0].keys())[0]


@vcr.use_cassette("test/vcr_cassettes/test_count_publishingcountries.yaml")
def test_count_publishingcountries():
    "occurrences.count_publishingcountries - basic test"
    res = occurrences.count_publishingcountries(country="DE")
    assert dict == res.__class__
    assert str == str(list(res.keys())[0]).__class__
    assert int == list(res.values())[0].__class__



import os
import hashlib
import re

import matplotlib.pyplot as plt
import matplotlib.image as mpimg
from appdirs import user_cache_dir

from pygbif.gbifutils import requests_argset, has, gbif_GET_map


def map(
    source="density",
    z=0,
    x=0,
    y=0,
    format="@1x.png",
    srs="EPSG:4326",
    bin=None,
    hexPerTile=None,
    style="classic.point",
    taxonKey=None,
    country=None,
    publishingCountry=None,
    publisher=None,
    datasetKey=None,
    year=None,
    basisOfRecord=None,
    **kwargs
):
    """
    GBIF maps API

    :param source: [str] Either ``density`` for fast, precalculated tiles,
        or ``adhoc`` for any search
    :param z: [str] zoom level
    :param x: [str] longitude
    :param y: [str] latitude
    :param format: [str] format of returned data. One of:

      - ``.mvt`` - vector tile
      - ``@Hx.png`` - 256px raster tile (for legacy clients)
      - ``@1x.png`` - 512px raster tile, @2x.png for a 1024px raster tile
      - ``@2x.png`` - 1024px raster tile
      - ``@3x.png`` - 2048px raster tile
      - ``@4x.png`` - 4096px raster tile

    :param srs: [str] Spatial reference system. One of:

      - ``EPSG:3857`` (Web Mercator)
      - ``EPSG:4326`` (WGS84 plate caree)
      - ``EPSG:3575`` (Arctic LAEA)
      - ``EPSG:3031`` (Antarctic stereographic)

    :param bin: [str] square or hex to aggregate occurrence counts into
        squares or hexagons. Points by default.
    :param hexPerTile: [str] sets the size of the hexagons (the number 
        horizontally across a tile)
    :param squareSize: [str] sets the size of the squares. Choose a factor 
        of 4096 so they tessalate correctly: probably from 8, 16, 32, 64, 
        128, 256, 512.
    :param style: [str] for raster tiles, choose from the available styles.
        Defaults to classic.point.
    :param taxonKey: [int] A GBIF occurrence identifier
    :param datasetKey: [str] The occurrence dataset key (a uuid)
    :param country: [str] The 2-letter country code (as per ISO-3166-1) of 
        the country in which the occurrence was recorded. See here
        http://en.wikipedia.org/wiki/ISO_3166-1_alpha-2
    :param basisOfRecord: [str] Basis of record, as defined in the BasisOfRecord enum 
        http://gbif.github.io/gbif-api/apidocs/org/gbif/api/vocabulary/BasisOfRecord.html
        Acceptable values are

       - ``FOSSIL_SPECIMEN`` An occurrence record describing a fossilized specimen.
       - ``HUMAN_OBSERVATION`` An occurrence record describing an observation made by one or more people.
       - ``LIVING_SPECIMEN`` An occurrence record describing a living specimen.
       - ``MACHINE_OBSERVATION`` An occurrence record describing an observation made by a machine.
       - ``MATERIAL_CITATION`` An occurrence record based on a reference to a scholarly publication.
       - ``OBSERVATION`` An occurrence record describing an observation.
       - ``OCCURRENCE`` An existence of an organism at a particular place and time. No more specific basis.
       - ``PRESERVED_SPECIMEN`` An occurrence record describing a preserved specimen.

    :param year: [int] The 4 digit year. A year of 98 will be interpreted as 
        AD 98. Supports range queries, smaller,larger (e.g., ``1990,1991``, 
        whereas ``1991,1990`` wouldn't work)
    :param publishingCountry: [str] The 2-letter country code (as per 
        ISO-3166-1) of the country in which the occurrence was recorded.

    :return: An object of class GbifMap

    For mvt format, see https://github.com/tilezen/mapbox-vector-tile to 
    decode, and example below

    Usage::

        from pygbif import maps
        out = maps.map(taxonKey = 2435098)
        out.response
        out.path
        out.img
        out.plot()

        out = maps.map(taxonKey = 2480498, year = range(2008, 2011+1))
        out.response
        out.path
        out.img
        out.plot()
        
        # srs
        maps.map(taxonKey = 2480498, year = 2010, srs = "EPSG:3857")
        # bin
        maps.map(taxonKey = 212, year = 1998, bin = "hex",
           hexPerTile = 30, style = "classic-noborder.poly")
        # style
        maps.map(taxonKey = 2480498, style = "purpleYellow.point").plot()
        # basisOfRecord
        maps.map(taxonKey = 2480498, year = 2010,
          basisOfRecord = "HUMAN_OBSERVATION", bin = "hex", 
          hexPerTile = 500).plot()
        maps.map(taxonKey = 2480498, year = 2010, 
          basisOfRecord = ["HUMAN_OBSERVATION", "LIVING_SPECIMEN"],
          hexPerTile = 500, bin = "hex").plot()

        # map vector tiles, gives back raw bytes
        from pygbif import maps
        x = maps.map(taxonKey = 2480498, year = 2010,
          format = ".mvt")
        x.response
        x.path
        x.img # None
        import mapbox_vector_tile
        mapbox_vector_tile.decode(x.response.content)
    """
    if format not in [".mvt", "@Hx.png", "@1x.png", "@2x.png", "@3x.png", "@4x.png"]:
        raise ValueError("'format' not in allowed set, see docs")
    if source not in ["density", "adhoc"]:
        raise ValueError("'source' not in allowed set, see docs")
    if srs not in ["EPSG:3857", "EPSG:4326", "EPSG:3575", "EPSG:3031"]:
        raise ValueError("'srs' not in allowed set, see docs")
    if bin is not None:
        if bin not in ["square", "hex"]:
            raise ValueError("'bin' not in allowed set, see docs")
    if style is not None:
        if style not in map_styles:
            raise ValueError("'style' not in allowed set, see docs")

    maps_baseurl = "https://api.gbif.org"
    url = maps_baseurl + "/v2/map/occurrence/%s/%s/%s/%s%s"
    url = url % (source, z, x, y, format)
    year = __handle_year(year)
    basisOfRecord = __handle_bor(basisOfRecord)
    args = {
        "srs": srs,
        "bin": bin,
        "hexPerTile": hexPerTile,
        "style": style,
        "taxonKey": taxonKey,
        "country": country,
        "publishingCountry": publishingCountry,
        "publisher": publisher,
        "datasetKey": datasetKey,
        "year": year,
        "basisOfRecord": basisOfRecord,
    }
    kw = {key: kwargs[key] for key in kwargs if key not in requests_argset}
    if kw is not None:
        xx = dict(zip([re.sub("_", ".", x) for x in kw.keys()], kw.values()))
        args.update(xx)
    kwargs = {key: kwargs[key] for key in kwargs if key in requests_argset}
    ctype = "image/png" if has(format, "png") else "application/x-protobuf"
    out = gbif_GET_map(url, args, ctype, **kwargs)
    # return out
    return GbifMap(out)


class GbifMap(object):
    """
    GbifMap response class

    contains:
    
    - response: the response from the requests library
    - path: the path to the image
    - img: the image data, of class matplotlib AxesImage
    - plot(): a method to plot the image with matplotlib
    """

    def __init__(self, x):
        super(GbifMap, self).__init__()
        self.response = x
        self.path = self.__make_path()
        self.__write_file()
        if has(self.response.headers["Content-Type"], "png"):
            self.img = self.__prep_plot()
        else:
            self.img = None

    def __write_file(self):
        fh = open(self.path, "wb")
        fh.write(self.response.content)
        fh.close()

    def __prep_plot(self):
        img = mpimg.imread(self.path)
        return plt.imshow(img)

    def __make_path(self):
        uu = hashlib.sha256(self.response.content).hexdigest()
        base_path = user_cache_dir("python/pygbif")
        if not os.path.exists(base_path):
            os.makedirs(base_path)
        file_ext = (
            ".png" if has(self.response.headers["Content-Type"], "png") else ".mvt"
        )
        path = base_path + "/" + uu + file_ext
        return path

    def plot(self):
        plt.show()


def __handle_year(year):
    if year is not None:
        if isinstance(year, int):
            year = [year]
        if isinstance(year, range):
            year = list(year)
        bools = [w in range(0, 2200) for w in year]
        if not all(bools):
            raise ValueError("one or more 'year' values not in acceptable set")
        if len(year) > 1:
            year = ",".join([str(min(year)), str(max(year))])
        else:
            year = str(year[0])
        return year


def __handle_bor(basisOfRecord):
    if basisOfRecord is not None:
        if isinstance(basisOfRecord, str):
            basisOfRecord = [basisOfRecord]
        bools = [w in basis_of_record_values for w in basisOfRecord]
        if not all(bools):
            raise ValueError("one or more 'basisOfRecord' values not in acceptable set")
        return basisOfRecord


map_styles = [
    "purpleHeat.point",
    "blueHeat.point",
    "orangeHeat.point",
    "greenHeat.point",
    "classic.point",
    "purpleYellow.point",
    "fire.point",
    "glacier.point",
    "classic.poly",
    "classic-noborder.poly",
    "purpleYellow.poly",
    "purpleYellow-noborder.poly",
    "green.poly",
    "green2.poly",
    "iNaturalist.poly",
    "purpleWhite.poly",
    "red.poly",
    "blue.marker",
    "orange.marker",
    "outline.poly",
]

basis_of_record_values = [
    "OBSERVATION",
    "HUMAN_OBSERVATION",
    "MACHINE_OBSERVATION",
    "MATERIAL_SAMPLE",
    "PRESERVED_SPECIMEN",
    "FOSSIL_SPECIMEN",
    "LIVING_SPECIMEN",
    "LITERATURE",
    "UNKNOWN",
]



from pygbif.gbifutils import gbif_baseurl, bool2str, requests_argset, gbif_GET
import re

def search(
    q=None,
    type=None,
    institutionalGovernance=None,
    disciplines=None,
    name=None,
    fuzzyName=None,
    numberSpecimens=None,
    occurrenceCount=None,
    typeSpecimenCount=None,
    sourceId=None,
    source=None,
    code=None,
    alternativeCode=None,
    contact=None,
    institutionKey=None,
    country=None,
    city=None,
    gbifRegion=None,
    machineTagNamespace=None,
    machineTagName=None,
    machineTagValue=None,
    identifier=None,
    identifierType=None,
    active=None,
    displayOnNHCPortal=None,
    masterSourceType=None,
    replacedBy=None,
    sortBy=None,
    sortOrder=None,
    offset=None,
    limit=None,
    **kwargs
    ): 
    """
    Search for institutions in GRSciColl.

    :param q: [str] Simple full text search parameter. The value for this parameter can be a simple word or a phrase. Wildcards are not supported
    :param type: [str] Type of a GrSciColl institution. Available values: BIOMEDICAL_RESEARCH_INSTITUTE, BOTANICAL_GARDEN, HERBARIUM, LIVING_ORGANISM_COLLECTION, MEDICAL_RESEARCH_INSTITUTE, MUSEUM, MUSEUM_HERBARIUM_PRIVATE_NON_PROFIT, OTHER_INSTITUTIONAL_TYPE, OTHER_TYPE_RESEARCH_INSTITUTION_BIOREPOSITORY, UNIVERSITY_COLLEGE, ZOO_AQUARIUM
    :param institutionalGovernance: [str] Institutional governance of a GrSciColl institution. Available values: ACADEMIC_FEDERAL, ACADEMIC_FOR_PROFIT, ACADEMIC_LOCAL, ACADEMIC_NON_PROFIT, ACADEMIC_STATE, FEDERAL, FOR_PROFIT, LOCAL, NON_PROFIT, OTHER, STATE
    :param disciplines: [str] Discipline of a GrSciColl institution. Check available values: https://techdocs.gbif.org/en/openapi/v1/registry#/Institutions/listInstitutions
    :param name: [str] Name of a GrSciColl institution or collection
    :param fuzzyName: [str] It searches by name fuzzily so the parameter doesn't have to be the exact name
    :param numberSpecimens: [str] Number of specimens. It supports ranges and a '*' can be used as a wildcard
    :param occurrenceCount: [str] Count of occurrences linked. It supports ranges and a '*' can be used as a wildcard
    :param typeSpecimenCount: [str] Count of type specimens linked. It supports ranges and a '*' can be used as a wildcard
    :param sourceId: [str] sourceId of MasterSourceMetadata
    :param source: [str] Source attribute of MasterSourceMetadata. Available values: DATASET, ORGANIZATION, IH_IRN
    :param code: [str] Code of a GrSciColl institution or collection
    :param alternativeCode: [str] Alternative code of a GrSciColl institution
    :param contact: [str] Filters collections and institutions whose contacts contain the person key specified
    :param institutionKey: [str] Keys of institutions to filter by
    :param country: [str] Filters by country given as a ISO 639-1 (2 letter) country code
    :param city: [str] Filters by the city of the address. It searches in both the physical and the mailing address
    :param gbifRegion: [str] Filters by a gbif region. Available values: AFRICA, ASIA, EUROPE, NORTH_AMERICA, OCEANIA, LATIN_AMERICA, ANTARCTICA
    :param machineTagNamespace: [str] Filters for entities with a machine tag in the specified namespace
    :param machineTagName: [str] Filters for entities with a machine tag with the specified name (use in combination with the machineTagNamespace parameter)
    :param machineTagValue: [str] Filters for entities with a machine tag with the specified value (use in combination with the machineTagNamespace and machineTagName parameters)
    :param identifier: [str] An identifier of the type given by the identifierType parameter, for example a DOI or UUID
    :param identifierType: [str] An identifier type for the identifier parameter. Available values: URL, LSID, HANDLER, DOI, UUID, FTP, URI, UNKNOWN, GBIF_PORTAL, GBIF_NODE, GBIF_PARTICIPANT, GRSCICOLL_ID, GRSCICOLL_URI, IH_IRN, ROR, GRID, CITES, SYMBIOTA_UUID, WIKIDATA, NCBI_BIOCOLLECTION, ISIL, CLB_DATASET_KEY
    :param active: [bool] Active status of a GrSciColl institution or collection
    :param displayOnNHCPortal: [bool] Flag to show this record in the NHC portal
    :param masterSourceType: [str] The master source type of a GRSciColl institution or collection. Available values: GRSCICOLL, GBIF_REGISTRY, IH
    :param replacedBy: [str] Key of the entity that replaced another entity
    :param sortBy: [str] Field to sort the results by. It only supports the fields contained in the enum. Available values: NUMBER_SPECIMENS
    :param sortOrder: [str] Sort order to use with the sortBy parameter. Available values: ASC, DESC
    :param offset: [int] Determines the offset for the search results
    :param limit: [int] Controls the number of results in the page. Default 20
    :param kwargs: Further named arguments passed on to requests.get

    :return: A dictionary

    Usage::
        from pygbif import institution as inst

        inst.search(q="Kansas",limit=1)
        inst.search(numberSpecimens = "1000,*",limit=2)
        inst.search(source = "IH_IRN") 
        inst.search(country = ["US","GB"])
        inst.search(typeSpecimenCount = "10,100")
    """
    url = gbif_baseurl + "grscicoll/institution"
    args = {
        "q": q,
        "type": type,
        "institutionalGovernance": institutionalGovernance,
        "disciplines": disciplines,
        "name": name,
        "fuzzyName": fuzzyName,
        "numberSpecimens": numberSpecimens,
        "occurrenceCount": occurrenceCount,
        "typeSpecimenCount": typeSpecimenCount,
        "sourceId": sourceId,
        "source": source,
        "code": code,
        "alternativeCode": alternativeCode,
        "contact": contact,
        "institutionKey": institutionKey,
        "country": country,
        "city": city,
        "gbifRegion": gbifRegion,
        "machineTagNamespace": machineTagNamespace,
        "machineTagName": machineTagName,
        "machineTagValue": machineTagValue,
        "identifier": identifier,
        "identifierType": identifierType,
        "active": active,
        "displayOnNHCPortal": displayOnNHCPortal,
        "masterSourceType": masterSourceType,
        "replacedBy": replacedBy,
        "sortBy": sortBy,
        "sortOrder": sortOrder,
        "offset": offset,
        "limit": limit, 
    }
    
    gbif_kwargs = {key: kwargs[key] for key in kwargs if key not in requests_argset}
    if gbif_kwargs is not None:
        xx = dict(
            zip([re.sub("_", ".", x) for x in gbif_kwargs.keys()], gbif_kwargs.values())
        )
        args.update(xx)
    kwargs = {key: kwargs[key] for key in kwargs if key in requests_argset}
    out = gbif_GET(url, args, **kwargs)
    return out



"""Tests for registry module - datasets method"""
import vcr
from pygbif import registry


@vcr.use_cassette("test/vcr_cassettes/test_datasets.yaml")
def test_datasets():
    "registry.datasets - basic test"
    res = registry.datasets()
    assert dict == res.__class__


@vcr.use_cassette("test/vcr_cassettes/test_datasets_limit.yaml")
def test_datasets_limit():
    "registry.datasets - limit param"
    res = registry.datasets(limit=1)
    assert dict == res.__class__
    assert 1 == len(res["results"])

    res = registry.datasets(limit=3)
    assert dict == res.__class__
    assert 3 == len(res["results"])


@vcr.use_cassette("test/vcr_cassettes/test_datasets_type.yaml")
def test_datasets_type():
    "registry.datasets - type param"
    res = registry.datasets(type="OCCURRENCE")
    vv = [x["type"] for x in res["results"]]
    assert dict == res.__class__
    assert 100 == len(res["results"])
    assert "OCCURRENCE" == list(set(vv))[0]



from pygbif.gbifutils import gbif_baseurl, gbif_POST


def name_parser(name, **kwargs):
    """
  Parse taxon names using the GBIF name parser

  :param name: [str] A character vector of scientific names. (required)

  reference: http://www.gbif.org/developer/species#parser

  Usage::

      from pygbif import species
      species.name_parser('x Agropogon littoralis')
      species.name_parser(['Arrhenatherum elatius var. elatius',
        'Secale cereale subsp. cereale', 'Secale cereale ssp. cereale',
        'Vanessa atalanta (Linnaeus, 1758)'])
  """
    url = gbif_baseurl + "parser/name"
    if name.__class__ == str:
        name = [name]
    return gbif_POST(url, name, **kwargs)



import re


def occ_issues_lookup(issue=None, code=None):
    """
    Lookup occurrence issue definitions and short codes

    :param issue: Full name of issue, e.g, CONTINENT_COUNTRY_MISMATCH
    :param code: an issue short code, e.g. ccm

    Usage
    pygbif.occ_issues_lookup(issue = 'CONTINENT_COUNTRY_MISMATCH')
    pygbif.occ_issues_lookup(issue = 'MULTIMEDIA_DATE_INVALID')
    pygbif.occ_issues_lookup(issue = 'ZERO_COORDINATE')
    pygbif.occ_issues_lookup(code = 'cdiv')
    """
    if code is None:
        bb = [trymatch(issue, x) for x in gbifissues["issue"]]
        tmp = filter(None, bb)
    else:
        bb = [trymatch(code, x) for x in gbifissues["code"]]
        tmp = filter(None, bb)
    return tmp


def trymatch(pattern, string):
    temp = re.match(pattern, string)
    if temp is None:
        return None
    else:
        return temp.string


gbifissues = {
    "code": [
        "bri",
        "ccm",
        "cdc",
        "conti",
        "cdiv",
        "cdout",
        "cdrep",
        "cdrepf",
        "cdreps",
        "cdround",
        "cucdmis",
        "cudc",
        "cuiv",
        "cum",
        "depmms",
        "depnn",
        "depnmet",
        "depunl",
        "elmms",
        "elnn",
        "elnmet",
        "elunl",
        "gass84",
        "gdativ",
        "iddativ",
        "iddatunl",
        "mdativ",
        "mdatunl",
        "muldativ",
        "muluriiv",
        "preneglat",
        "preneglon",
        "preswcd",
        "rdativ",
        "rdatm",
        "rdatunl",
        "refuriiv",
        "txmatfuz",
        "txmathi",
        "txmatnon",
        "typstativ",
        "zerocd",
    ],
    "issue": [
        "BASIS_OF_RECORD_INVALID",
        "CONTINENT_COUNTRY_MISMATCH",
        "CONTINENT_DERIVED_FROM_COORDINATES",
        "CONTINENT_INVALID",
        "COORDINATE_INVALID",
        "COORDINATE_OUT_OF_RANGE",
        "COORDINATE_REPROJECTED",
        "COORDINATE_REPROJECTION_FAILED",
        "COORDINATE_REPROJECTION_SUSPICIOUS",
        "COORDINATE_ROUNDED",
        "COUNTRY_COORDINATE_MISMATCH",
        "COUNTRY_DERIVED_FROM_COORDINATES",
        "COUNTRY_INVALID",
        "COUNTRY_MISMATCH",
        "DEPTH_MIN_MAX_SWAPPED",
        "DEPTH_NON_NUMERIC",
        "DEPTH_NOT_METRIC",
        "DEPTH_UNLIKELY",
        "ELEVATION_MIN_MAX_SWAPPED",
        "ELEVATION_NON_NUMERIC",
        "ELEVATION_NOT_METRIC",
        "ELEVATION_UNLIKELY",
        "GEODETIC_DATUM_ASSUMED_WGS84",
        "GEODETIC_DATUM_INVALID",
        "IDENTIFIED_DATE_INVALID",
        "IDENTIFIED_DATE_UNLIKELY",
        "MODIFIED_DATE_INVALID",
        "MODIFIED_DATE_UNLIKELY",
        "MULTIMEDIA_DATE_INVALID",
        "MULTIMEDIA_URI_INVALID",
        "PRESUMED_NEGATED_LATITUDE",
        "PRESUMED_NEGATED_LONGITUDE",
        "PRESUMED_SWAPPED_COORDINATE",
        "RECORDED_DATE_INVALID",
        "RECORDED_DATE_MISMATCH",
        "RECORDED_DATE_UNLIKELY",
        "REFERENCES_URI_INVALID",
        "TAXON_MATCH_FUZZY",
        "TAXON_MATCH_HIGHERRANK",
        "TAXON_MATCH_NONE",
        "TYPE_STATUS_INVALID",
        "ZERO_COORDINATE",
    ],
    "description": [
        "The given basis of record is impossible to interpret or seriously different from the recommended vocabulary.",
        "The interpreted continent and country do not match up.",
        "The interpreted continent is based on the coordinates, not the verbatim string information.",
        "Uninterpretable continent values found.",
        "Coordinate value given in some form but GBIF is unable to interpret it.",
        "Coordinate has invalid lat/lon values out of their decimal max range.",
        "The original coordinate was successfully reprojected from a different geodetic datum to WGS84.",
        "The given decimal latitude and longitude could not be reprojected to WGS84 based on the provided datum.",
        "Indicates successful coordinate reprojection according to provided datum, but which results in a datum shift larger than 0.1 decimal degrees.",
        "Original coordinate modified by rounding to 5 decimals.",
        "The interpreted occurrence coordinates fall outside of the indicated country.",
        "The interpreted country is based on the coordinates, not the verbatim string information.",
        "Uninterpretable country values found.",
        "Interpreted country for dwc:country and dwc:countryCode contradict each other.",
        "Set if supplied min>max",
        "Set if depth is a non numeric value",
        "Set if supplied depth is not given in the metric system, for example using feet instead of meters",
        "Set if depth is larger than 11.000m or negative.",
        "Set if supplied min > max elevation",
        "Set if elevation is a non numeric value",
        "Set if supplied elevation is not given in the metric system, for example using feet instead of meters",
        "Set if elevation is above the troposphere (17km) or below 11km (Mariana Trench).",
        "Indicating that the interpreted coordinates assume they are based on WGS84 datum as the datum was either not indicated or interpretable.",
        "The geodetic datum given could not be interpreted.",
        "The date given for dwc:dateIdentified is invalid and cant be interpreted at all.",
        "The date given for dwc:dateIdentified is in the future or before Linnean times (1700).",
        "A (partial) invalid date is given for dc:modified, such as a non existing date, invalid zero month, etc.",
        "The date given for dc:modified is in the future or predates unix time (1970).",
        "An invalid date is given for dc:created of a multimedia object.",
        "An invalid uri is given for a multimedia object.",
        "Latitude appears to be negated, e.g. 32.3 instead of -32.3",
        "Longitude appears to be negated, e.g. 32.3 instead of -32.3",
        "Latitude and longitude appear to be swapped.",
        "A (partial) invalid date is given, such as a non existing date, invalid zero month, etc.",
        "The recording date specified as the eventDate string and the individual year, month, day are contradicting.",
        "The recording date is highly unlikely, falling either into the future or represents a very old date before 1600 that predates modern taxonomy.",
        "An invalid uri is given for dc:references.",
        "Matching to the taxonomic backbone can only be done using a fuzzy, non exact match.",
        "Matching to the taxonomic backbone can only be done on a higher rank and not the scientific name.",
        "Matching to the taxonomic backbone cannot be done cause there was no match at all or several matches with too little information to keep them apart (homonyms).",
        "The given type status is impossible to interpret or seriously different from the recommended vocabulary.",
        "Coordinate is the exact 0/0 coordinate, often indicating a bad null coordinate.",
    ],
}



from pygbif import occurrences
import vcr

@vcr.use_cassette('test/vcr_cassettes/test-occurrences-download_sql.yaml')
def test_download_sql():
    """basic test of the download_sql function"""
    out = occurrences.download_sql("SELECT gbifid,publishingCountry FROM occurrence WHERE publishingCountry='BE'")
    assert "str" == out.__class__.__name__
    assert 23 == len(out)





from pygbif.gbifutils import (
    check_data,
    stop,
    gbif_baseurl,
    gbif_GET,
    get_meta,
    parse_results,
    len2,
)


def networks(
    data="all",
    uuid=None,
    q=None,
    identifier=None,
    identifierType=None,
    limit=100,
    offset=None,
    **kwargs
):
    """
  Networks metadata.

  Note: there's only 1 network now, so there's not a lot you can do with this method.

  :param data: [str] The type of data to get. Default: ``all``
  :param uuid: [str] UUID of the data network provider. This must be specified if data
     is anything other than ``all``.
  :param q: [str] Query networks. Only used when ``data = 'all'``. Ignored otherwise.
  :param identifier: [fixnum] The value for this parameter can be a simple string or integer,
      e.g. identifier=120
  :param identifierType: [str] Used in combination with the identifier parameter to filter
      identifiers by identifier type: ``DOI``, ``FTP``, ``GBIF_NODE``, ``GBIF_PARTICIPANT``,
      ``GBIF_PORTAL``, ``HANDLER``, ``LSID``, ``UNKNOWN``, ``URI``, ``URL``, ``UUID``
  :param limit: [int] Number of results to return. Default: ``100``
  :param offset: [int] Record to start at. Default: ``0``

  :return: A dictionary

  References: http://www.gbif.org/developer/registry#networks

  Usage::

      from pygbif import registry
      registry.networks(limit=1)
      registry.networks(uuid='2b7c7b4f-4d4f-40d3-94de-c28b6fa054a6')
  """
    args = {
        "q": q,
        "limit": limit,
        "offset": offset,
        "identifier": identifier,
        "identifierType": identifierType,
    }
    data_choices = [
        "all",
        "contact",
        "endpoint",
        "identifier",
        "tag",
        "machineTag",
        "comment",
        "constituents",
    ]
    check_data(data, data_choices)

    def getdata(x, uuid, args, **kwargs):
        if x != "all" and uuid is None:
            stop('You must specify a uuid if data does not equal "all"')

        if uuid is None:
            url = gbif_baseurl + "network"
        else:
            if x == "all":
                url = gbif_baseurl + "network/" + uuid
            else:
                url = gbif_baseurl + "network/" + uuid + "/" + x

        res = gbif_GET(url, args, **kwargs)
        return {"meta": get_meta(res), "data": parse_results(res, uuid)}

    if len2(data) == 1:
        return getdata(data, uuid, args, **kwargs)
    else:
        return [getdata(x, uuid, args, **kwargs) for x in data]



"""tests for literature search"""
import vcr
from pygbif import literature

keyz = ["count", "facets", "results", "endOfRecords", "limit", "offset"]

@vcr.use_cassette("test/vcr_cassettes/test_literature_search.yaml")
def test_search():
    "literature.search - basic test"
    res = literature.search(limit=10)
    assert "dict" == res.__class__.__name__
    assert 6 == len(res)
    assert sorted(keyz) == sorted(res.keys())
    assert 10 == len(res["results"])
    assert res["count"] >= 48132 # unlikely to go down 

@vcr.use_cassette("test/vcr_cassettes/test_literature_search_key.yaml")
def test_search_key():
    "literature.search - key"
    res = literature.search(gbifDownloadKey="0235283-220831081235567")
    assert "dict" == res.__class__.__name__
    assert 6 == len(res)
    assert sorted(keyz) == sorted(res.keys())
    assert 1 == len(res["results"])
    assert 2024 == res["results"][0]["year"]
    assert ["0235283-220831081235567","0026791-240906103802322"] == res["results"][0]["gbifDownloadKey"]

@vcr.use_cassette("test/vcr_cassettes/test_literature_search_year.yaml")
def test_search_year():
    "literature.search - year"
    res = literature.search(year="2010,2024", limit=10)
    assert "dict" == res.__class__.__name__
    assert 6 == len(res)
    assert sorted(keyz) == sorted(res.keys())
    assert 10 == len(res["results"])
    for result in res["results"]:
        assert 2010 <= result["year"] <= 2024

@vcr.use_cassette("test/vcr_cassettes/test_literature_search_facet.yaml")
def test_search_facet():
    "literature.search - facet"
    res = literature.search(facet="year", limit=10)
    assert "dict" == res.__class__.__name__
    assert 6 == len(res)
    assert sorted(keyz) == sorted(res.keys())
    assert 10 == len(res["results"])
    assert "list" == res["facets"].__class__.__name__
    assert 2 == len(res["facets"][0]["counts"][0])
    assert set(['name', 'count']) == set(res["facets"][0]["counts"][0].keys())




"""Tests for occurrences module - download_get methods"""
from pygbif import occurrences as occ


def test_download_get():
    "occurrences.download_get - basic test"
    key = "0089857-160910150852091"
    res = occ.download_get(key)
    assert "dict" == res.__class__.__name__
    assert len(res) == 3
    assert key == res["key"]



"""Tests for occurrences module - download_cancle methods"""
from pygbif import occurrences as occ
import vcr
import os
import pytest

IN_GITHUB_ACTIONS = os.getenv("GITHUB_ACTIONS") == "true"
print(IN_GITHUB_ACTIONS)

@pytest.mark.skipif(IN_GITHUB_ACTIONS, reason="Test doesn't work well in Github Actions.")
@vcr.use_cassette("test/vcr_cassettes/test_download_cancel.yaml", filter_headers=["authorization"])
def test_download_cancel():
    "occurrences.download_cancel - basic test"
    name_key = "789"  # Odonata
    res = occ.download("taxonKey = " + name_key)
    download_key = res[0]
    out = occ.download_cancel(download_key)
    assert True == out



import re

from pygbif.gbifutils import gbif_baseurl, bool2str, requests_argset, gbif_GET


def search(
    taxonKey=None,
    repatriated=None,
    kingdomKey=None,
    phylumKey=None,
    classKey=None,
    orderKey=None,
    familyKey=None,
    genusKey=None,
    subgenusKey=None,
    scientificName=None,
    country=None,
    publishingCountry=None,
    hasCoordinate=None,
    typeStatus=None,
    recordNumber=None,
    lastInterpreted=None,
    continent=None,
    geometry=None,
    recordedBy=None,
    recordedByID=None,
    identifiedByID=None,
    basisOfRecord=None,
    datasetKey=None,
    eventDate=None,
    catalogNumber=None,
    year=None,
    month=None,
    decimalLatitude=None,
    decimalLongitude=None,
    elevation=None,
    depth=None,
    institutionCode=None,
    collectionCode=None,
    hasGeospatialIssue=None,
    issue=None,
    q=None,
    spellCheck=None,
    mediatype=None,
    limit=300,
    offset=0,
    establishmentMeans=None,
    facet=None,
    facetMincount=None,
    facetMultiselect=None,
    **kwargs
):
    """
    Search GBIF occurrences

    :param taxonKey: [int] A GBIF occurrence identifier
    :param q: [str] Simple search parameter. The value for this parameter can be a simple word or a phrase.
    :param spellCheck: [bool] If ``True`` ask GBIF to check your spelling of the value passed to the ``search`` parameter.
        IMPORTANT: This only checks the input to the ``search`` parameter, and no others. Default: ``False``
    :param repatriated: [str] Searches for records whose publishing country is different to the country where the record was recorded in
    :param kingdomKey: [int] Kingdom classification key
    :param phylumKey: [int] Phylum classification key
    :param classKey: [int] Class classification key
    :param orderKey: [int] Order classification key
    :param familyKey: [int] Family classification key
    :param genusKey: [int] Genus classification key
    :param subgenusKey: [int] Subgenus classification key
    :param scientificName: [str] A scientific name from the GBIF backbone. All included and synonym taxa are included in the search.
    :param datasetKey: [str] The occurrence dataset key (a uuid)
    :param catalogNumber: [str] An identifier of any form assigned by the source within a physical collection or digital dataset for the record which may not unique, but should be fairly unique in combination with the institution and collection code.
    :param recordedBy: [str] The person who recorded the occurrence.
    :param recordedByID: [str] Identifier (e.g. ORCID) for the person who recorded the occurrence
    :param identifiedByID: [str] Identifier (e.g. ORCID) for the person who provided the taxonomic identification of the occurrence.
    :param collectionCode: [str] An identifier of any form assigned by the source to identify the physical collection or digital dataset uniquely within the text of an institution.
    :param institutionCode: [str] An identifier of any form assigned by the source to identify the institution the record belongs to. Not guaranteed to be que.
    :param country: [str] The 2-letter country code (as per ISO-3166-1) of the country in which the occurrence was recorded. See here http://en.wikipedia.org/wiki/ISO_3166-1_alpha-2
    :param basisOfRecord: [str] Basis of record, as defined in our BasisOfRecord enum here http://gbif.github.io/gbif-api/apidocs/org/gbif/api/vocabulary/BasisOfRecord.html Acceptable values are:

     - ``FOSSIL_SPECIMEN`` An occurrence record describing a fossilized specimen.
     - ``HUMAN_OBSERVATION`` An occurrence record describing an observation made by one or more people.
     - ``LIVING_SPECIMEN`` An occurrence record describing a living specimen.
     - ``MACHINE_OBSERVATION`` An occurrence record describing an observation made by a machine.
     - ``MATERIAL_CITATION`` An occurrence record based on a reference to a scholarly publication.
     - ``OBSERVATION`` An occurrence record describing an observation.
     - ``OCCURRENCE`` An existence of an organism at a particular place and time. No more specific basis.
     - ``PRESERVED_SPECIMEN`` An occurrence record describing a preserved specimen.

    :param eventDate: [date] Occurrence date in ISO 8601 format: yyyy, yyyy-MM, yyyy-MM-dd, or
       MM-dd. Supports range queries, smaller,larger (e.g., ``1990,1991``, whereas ``1991,1990``
       wouldn't work)
    :param year: [int] The 4 digit year. A year of 98 will be interpreted as AD 98. Supports range queries,
       smaller,larger (e.g., ``1990,1991``, whereas ``1991,1990`` wouldn't work)
    :param month: [int] The month of the year, starting with 1 for January. Supports range queries,
       smaller,larger (e.g., ``1,2``, whereas ``2,1`` wouldn't work)
    :param decimalLatitude: [float] Latitude in decimals between -90 and 90 based on WGS 84.
       Supports range queries, smaller,larger (e.g., ``25,30``, whereas ``30,25`` wouldn't work)
    :param decimalLongitude: [float] Longitude in decimals between -180 and 180 based on WGS 84.
       Supports range queries (e.g., ``-0.4,-0.2``, whereas ``-0.2,-0.4`` wouldn't work).
    :param publishingCountry: [str] The 2-letter country code (as per ISO-3166-1) of the
       country in which the occurrence was recorded.
    :param elevation: [int/str] Elevation in meters above sea level. Supports range queries, smaller,larger
       (e.g., ``5,30``, whereas ``30,5`` wouldn't work)
    :param depth: [int/str] Depth in meters relative to elevation. For example 10 meters below a
       lake surface with given elevation. Supports range queries, smaller,larger (e.g., ``5,30``,
       whereas ``30,5`` wouldn't work)
    :param geometry: [str] Searches for occurrences inside a polygon described in Well Known
       Text (WKT) format. A WKT shape written as either POINT, LINESTRING, LINEARRING
       POLYGON, or MULTIPOLYGON. Example of a polygon: ``((30.1 10.1, 20, 20 40, 40 40, 30.1 10.1))`` would be queried as http://bit.ly/1BzNwDq.
       Polygons must have counter-clockwise ordering of points.
    :param hasGeospatialIssue: [bool] Includes/excludes occurrence records which contain spatial
       issues (as determined in our record interpretation), i.e. ``hasGeospatialIssue=TRUE``
       returns only those records with spatial issues while ``hasGeospatialIssue=FALSE`` includes
       only records without spatial issues. The absence of this parameter returns any
       record with or without spatial issues.
    :param issue: [str] One or more of many possible issues with each occurrence record. See
       Details. Issues passed to this parameter filter results by the issue.
    :param hasCoordinate: [bool] Return only occurence records with lat/long data (``True``) or
       all records (``False``, default).
    :param typeStatus: [str] Type status of the specimen. One of many options. See ?typestatus
    :param recordNumber: [int] Number recorded by collector of the data, different from GBIF record
       number. See http://rs.tdwg.org/dwc/terms/#recordNumber} for more info
    :param lastInterpreted: [date] Date the record was last modified in GBIF, in ISO 8601 format:
       yyyy, yyyy-MM, yyyy-MM-dd, or MM-dd.  Supports range queries, smaller,larger (e.g.,
       ``1990,1991``, whereas ``1991,1990`` wouldn't work)
    :param continent: [str] Continent. One of ``africa``, ``antarctica``, ``asia``, ``europe``, ``north_america``
       (North America includes the Caribbean and reachies down and includes Panama), ``oceania``,
       or ``south_america``
    :param fields: [str] Default (``all``) returns all fields. ``minimal`` returns just taxon name,
       key, latitude, and longitude. Or specify each field you want returned by name, e.g.
       ``fields = ['name','latitude','elevation']``.
    :param mediatype: [str] Media type. Default is ``NULL``, so no filtering on mediatype. Options:
       ``NULL``, ``MovingImage``, ``Sound``, and ``StillImage``
    :param limit: [int] Number of results to return. Default: ``300``
    :param offset: [int] Record to start at. Default: ``0``
    :param facet: [str] a character vector of length 1 or greater
    :param establishmentMeans: [str] EstablishmentMeans, possible values include: INTRODUCED,
        INVASIVE, MANAGED, NATIVE, NATURALISED, UNCERTAIN
    :param facetMincount: [int] minimum number of records to be included in the faceting results
    :param facetMultiselect: [bool] Set to ``True`` to still return counts for values that are not currently
        filtered. See examples. Default: ``False``

    :return: A dictionary

    Usage::

        from pygbif import occurrences
        occurrences.search(taxonKey = 3329049)

        # Return 2 results, this is the default by the way
        occurrences.search(taxonKey=3329049, limit=2)

        # Instead of getting a taxon key first, you can search for a name directly
        # However, note that using this approach (with `scientificName="..."`)
        # you are getting synonyms too. The results for using `scientifcName` and
        # `taxonKey` parameters are the same in this case, but I wouldn't be surprised if for some
        # names they return different results
        occurrences.search(scientificName = 'Ursus americanus')
        from pygbif import species
        key = species.name_backbone(name = 'Ursus americanus', rank='species')['usageKey']
        occurrences.search(taxonKey = key)

        # Search by dataset key
        occurrences.search(datasetKey='7b5d6a48-f762-11e1-a439-00145eb45e9a', limit=20)

        # Search by catalog number
        occurrences.search(catalogNumber="49366", limit=20)
        # occurrences.search(catalogNumber=["49366","Bird.27847588"], limit=20)

        # Use paging parameters (limit and offset) to page. Note the different results
        # for the two queries below.
        occurrences.search(datasetKey='7b5d6a48-f762-11e1-a439-00145eb45e9a', offset=10, limit=5)
        occurrences.search(datasetKey='7b5d6a48-f762-11e1-a439-00145eb45e9a', offset=20, limit=5)

        # Many dataset keys
        # occurrences.search(datasetKey=["50c9509d-22c7-4a22-a47d-8c48425ef4a7", "7b5d6a48-f762-11e1-a439-00145eb45e9a"], limit=20)

        # Search by collector name
        res = occurrences.search(recordedBy="smith", limit=20)
        [ x['recordedBy'] for x in res['results'] ]

        # Many collector names
        # occurrences.search(recordedBy=["smith","BJ Stacey"], limit=20)
        
        # recordedByID
        occurrences.search(recordedByID="https://orcid.org/0000-0003-1691-239X", limit = 3)

        # identifiedByID
        occurrences.search(identifiedByID="https://orcid.org/0000-0003-1691-239X", limit = 3)

        # Search for many species
        splist = ['Cyanocitta stelleri', 'Junco hyemalis', 'Aix sponsa']
        keys = [ species.name_suggest(x)[0]['key'] for x in splist ]
        out = [ occurrences.search(taxonKey = x, limit=1) for x in keys ]
        [ x['results'][0]['speciesKey'] for x in out ]

        # Search - q parameter
        occurrences.search(q = "kingfisher", limit=20)
        ## spell check - only works with the `search` parameter
        ### spelled correctly - same result as above call
        occurrences.search(q = "kingfisher", limit=20, spellCheck = True)
        ### spelled incorrectly - stops with suggested spelling
        occurrences.search(q = "kajsdkla", limit=20, spellCheck = True)
        ### spelled incorrectly - stops with many suggested spellings
        ###   and number of results for each
        occurrences.search(q = "helir", limit=20, spellCheck = True)

        # Search on latitidue and longitude
        occurrences.search(decimalLatitude=50, decimalLongitude=10, limit=2)

        # Search on a bounding box
        ## in well known text format
        occurrences.search(geometry='POLYGON((30.1 10.1, 10 20, 20 40, 40 40, 30.1 10.1))', limit=20)
        from pygbif import species
        key = species.name_suggest(q='Aesculus hippocastanum')[0]['key']
        occurrences.search(taxonKey=key, geometry='POLYGON((30.1 10.1, 10 20, 20 40, 40 40, 30.1 10.1))', limit=20)
        ## multipolygon
        wkt = 'MULTIPOLYGON(((-123 38, -123 43, -116 43, -116 38, -123 38)),((-97 41, -97 45, -93 45, -93 41, -97 41)))'
        occurrences.search(geometry = wkt, limit = 20)

        # Search on country
        occurrences.search(country='US', limit=20)
        occurrences.search(country='FR', limit=20)
        occurrences.search(country='DE', limit=20)

        # Get only occurrences with lat/long data
        occurrences.search(taxonKey=key, hasCoordinate=True, limit=20)

        # Get only occurrences that were recorded as living specimens
        occurrences.search(taxonKey=key, basisOfRecord="LIVING_SPECIMEN", hasCoordinate=True, limit=20)

        # Get occurrences for a particular eventDate
        occurrences.search(taxonKey=key, eventDate="2013", limit=20)
        occurrences.search(taxonKey=key, year="2013", limit=20)
        occurrences.search(taxonKey=key, month="6", limit=20)

        # Get occurrences based on depth
        key = species.name_backbone(name='Salmo salar', kingdom='animals')['usageKey']
        occurrences.search(taxonKey=key, depth="5", limit=20)

        # Get occurrences based on elevation
        key = species.name_backbone(name='Puma concolor', kingdom='animals')['usageKey']
        occurrences.search(taxonKey=key, elevation=50, hasCoordinate=True, limit=20)

        # Get occurrences based on institutionCode
        occurrences.search(institutionCode="TLMF", limit=20)

        # Get occurrences based on collectionCode
        occurrences.search(collectionCode="Floristic Databases MV - Higher Plants", limit=20)

        # Get only those occurrences with spatial issues
        occurrences.search(taxonKey=key, hasGeospatialIssue=True, limit=20)

        # Search using a query string
        occurrences.search(q="kingfisher", limit=20)

        # Range queries
        ## See Detail for parameters that support range queries
        ### this is a range depth, with lower/upper limits in character string
        occurrences.search(depth='50,100')

        ## Range search with year
        occurrences.search(year='1999,2000', limit=20)

        ## Range search with latitude
        occurrences.search(decimalLatitude='29.59,29.6')

        # Search by specimen type status
        ## Look for possible values of the typeStatus parameter looking at the typestatus dataset
        occurrences.search(typeStatus = 'allotype')

        # Search by specimen record number
        ## This is the record number of the person/group that submitted the data, not GBIF's numbers
        ## You can see that many different groups have record number 1, so not super helpful
        occurrences.search(recordNumber = 1)

        # Search by last time interpreted: Date the record was last modified in GBIF
        ## The lastInterpreted parameter accepts ISO 8601 format dates, including
        ## yyyy, yyyy-MM, yyyy-MM-dd, or MM-dd. Range queries are accepted for lastInterpreted
        occurrences.search(lastInterpreted = '2014-04-01')

        # Search by continent
        ## One of africa, antarctica, asia, europe, north_america, oceania, or south_america
        occurrences.search(continent = 'south_america')
        occurrences.search(continent = 'africa')
        occurrences.search(continent = 'oceania')
        occurrences.search(continent = 'antarctica')

        # Search for occurrences with images
        occurrences.search(mediatype = 'StillImage')
        occurrences.search(mediatype = 'MovingImage')
        x = occurrences.search(mediatype = 'Sound')
        [z['media'] for z in x['results']]

        # Query based on issues
        occurrences.search(taxonKey=1, issue='DEPTH_UNLIKELY')
        occurrences.search(taxonKey=1, issue=['DEPTH_UNLIKELY','COORDINATE_ROUNDED'])
        # Show all records in the Arizona State Lichen Collection that cant be matched to the GBIF
        # backbone properly:
        occurrences.search(datasetKey='84c0e1a0-f762-11e1-a439-00145eb45e9a', issue=['TAXON_MATCH_NONE','TAXON_MATCH_HIGHERRANK'])

        # If you pass in an invalid polygon you get hopefully informative errors
        ### the WKT string is fine, but GBIF says bad polygon
        wkt = 'POLYGON((-178.59375 64.83258989321493,-165.9375 59.24622380205539,
        -147.3046875 59.065977905449806,-130.78125 51.04484764446178,-125.859375 36.70806354647625,
        -112.1484375 23.367471303759686,-105.1171875 16.093320185359257,-86.8359375 9.23767076398516,
        -82.96875 2.9485268155066175,-82.6171875 -14.812060061226388,-74.8828125 -18.849111862023985,
        -77.34375 -47.661687803329166,-84.375 -49.975955187343295,174.7265625 -50.649460483096114,
        179.296875 -42.19189902447192,-176.8359375 -35.634976650677295,176.8359375 -31.835565983656227,
        163.4765625 -6.528187613695323,152.578125 1.894796132058301,135.703125 4.702353722559447,
        127.96875 15.077427674847987,127.96875 23.689804541429606,139.921875 32.06861069132688,
        149.4140625 42.65416193033991,159.2578125 48.3160811030533,168.3984375 57.019804336633165,
        178.2421875 59.95776046458139,-179.6484375 61.16708631440347,-178.59375 64.83258989321493))'
        occurrences.search(geometry = wkt)

        # Faceting
        ## return no occurrence records with limit=0
        x = occurrences.search(facet = "country", limit = 0)
        x['facets']

        ## also return occurrence records
        x = occurrences.search(facet = "establishmentMeans", limit = 10)
        x['facets']
        x['results']

        ## multiple facet variables
        x = occurrences.search(facet = ["country", "basisOfRecord"], limit = 10)
        x['results']
        x['facets']
        x['facets']['country']
        x['facets']['basisOfRecord']
        x['facets']['basisOfRecord']['count']

        ## set a minimum facet count
        x = occurrences.search(facet = "country", facetMincount = 30000000L, limit = 0)
        x['facets']

        ## paging per each faceted variable
        ### do so by passing in variables like "country" + "_facetLimit" = "country_facetLimit"
        ### or "country" + "_facetOffset" = "country_facetOffset"
        x = occurrences.search(
          facet = ["country", "basisOfRecord", "hasCoordinate"],
          country_facetLimit = 3,
          basisOfRecord_facetLimit = 6,
          limit = 0
        )
        x['facets']

        # requests package options
        ## There's an acceptable set of requests options (['timeout', 'cookies', 'auth',
        ## 'allow_redirects', 'proxies', 'verify', 'stream', 'cert']) you can pass
        ## in via **kwargs, e.g., set a timeout
        x = occurrences.search(timeout = 1)
    """
    url = gbif_baseurl + "occurrence/search"
    args = {
        "taxonKey": taxonKey,
        "repatriated": repatriated,
        "kingdomKey": kingdomKey,
        "phylumKey": phylumKey,
        "classKey": classKey,
        "orderKey": orderKey,
        "familyKey": familyKey,
        "genusKey": genusKey,
        "subgenusKey": subgenusKey,
        "scientificName": scientificName,
        "country": country,
        "publishingCountry": publishingCountry,
        "hasCoordinate": bool2str(hasCoordinate),
        "typeStatus": typeStatus,
        "recordNumber": recordNumber,
        "lastInterpreted": lastInterpreted,
        "continent": continent,
        "geometry": geometry,
        "recordedBy": recordedBy,
        "recordedByID": recordedByID,
        "identifiedByID": identifiedByID,
        "basisOfRecord": basisOfRecord,
        "datasetKey": datasetKey,
        "eventDate": eventDate,
        "catalogNumber": catalogNumber,
        "year": year,
        "month": month,
        "decimalLatitude": decimalLatitude,
        "decimalLongitude": decimalLongitude,
        "elevation": elevation,
        "depth": depth,
        "institutionCode": institutionCode,
        "collectionCode": collectionCode,
        "hasGeospatialIssue": bool2str(hasGeospatialIssue),
        "issue": issue,
        "q": q,
        "spellCheck": bool2str(spellCheck),
        "mediatype": mediatype,
        "limit": limit,
        "offset": offset,
        "establishmentMeans": establishmentMeans,
        "facetMincount": facetMincount,
        "facet": facet,
        "facetMultiselect": bool2str(facetMultiselect),
    }
    gbif_kwargs = {key: kwargs[key] for key in kwargs if key not in requests_argset}
    if gbif_kwargs is not None:
        xx = dict(
            zip([re.sub("_", ".", x) for x in gbif_kwargs.keys()], gbif_kwargs.values())
        )
        args.update(xx)
    kwargs = {key: kwargs[key] for key in kwargs if key in requests_argset}
    out = gbif_GET(url, args, **kwargs)
    return out



from pygbif.gbifutils import gbif_baseurl, bool2str, gbif_GET, check_param_lens


def count(
    taxonKey=None,
    basisOfRecord=None,
    country=None,
    isGeoreferenced=None,
    datasetKey=None,
    publishingCountry=None,
    typeStatus=None,
    issue=None,
    year=None,
    **kwargs
):
    """
    Returns occurrence counts for a predefined set of dimensions

    For all parameters below, only one value allowed per function call.
    See :func:`~occurrences.search` for passing more than one value
    per parameter.

    :param taxonKey: [int] A GBIF occurrence identifier
    :param basisOfRecord: [str] A GBIF occurrence identifier
    :param country: [str] A GBIF occurrence identifier
    :param isGeoreferenced: [bool] A GBIF occurrence identifier
    :param datasetKey: [str] A GBIF occurrence identifier
    :param publishingCountry: [str] A GBIF occurrence identifier
    :param typeStatus: [str] A GBIF occurrence identifier
    :param issue: [str] A GBIF occurrence identifier
    :param year: [int] A GBIF occurrence identifier

    :return: dict

    Usage::

        from pygbif import occurrences
        occurrences.count(taxonKey = 3329049)
        occurrences.count(country = 'CA')
        occurrences.count(isGeoreferenced = True)
        occurrences.count(basisOfRecord = 'OBSERVATION')
    """
    check_param_lens(
        taxonKey=taxonKey,
        basisOfRecord=basisOfRecord,
        country=country,
        isGeoreferenced=isGeoreferenced,
        datasetKey=datasetKey,
        publishingCountry=publishingCountry,
        typeStatus=typeStatus,
        issue=issue,
        year=year,
    )
    url = gbif_baseurl + "occurrence/count"
    isGeoreferenced = bool2str(isGeoreferenced)
    out = gbif_GET(
        url,
        {
            "taxonKey": taxonKey,
            "basisOfRecord": basisOfRecord,
            "country": country,
            "isGeoreferenced": isGeoreferenced,
            "datasetKey": datasetKey,
            "publishingCountry": publishingCountry,
            "typeStatus": typeStatus,
            "issue": issue,
            "year": year,
        },
        **kwargs
    )
    return out


def count_basisofrecord(**kwargs):
    """
    Lists occurrence counts by basis of record.

    :return: dict

    Usage::

            from pygbif import occurrences
            occurrences.count_basisofrecord()
    """
    url = gbif_baseurl + "occurrence/counts/basisOfRecord"
    out = gbif_GET(url, {}, **kwargs)
    return out


def count_year(year, **kwargs):
    """
    Lists occurrence counts by year

    :param year: [int] year range, e.g., ``1990,2000``. Does not support ranges like ``asterisk,2010``

    :return: dict

    Usage::

            from pygbif import occurrences
            occurrences.count_year(year = '1990,2000')
    """
    url = gbif_baseurl + "occurrence/counts/year"
    out = gbif_GET(url, {"year": year}, **kwargs)
    return out


def count_datasets(taxonKey=None, country=None, **kwargs):
    """
    Lists occurrence counts for datasets that cover a given taxon or country

    :param taxonKey: [int] Taxon key
    :param country: [str] A country, two letter code

    :return: dict

    Usage::

            from pygbif import occurrences
            occurrences.count_datasets(country = "DE")
    """
    url = gbif_baseurl + "occurrence/counts/datasets"
    out = gbif_GET(url, {"taxonKey": taxonKey, "country": country}, **kwargs)
    return out


def count_countries(publishingCountry, **kwargs):
    """
    Lists occurrence counts for all countries covered by the data published by the given country

    :param publishingCountry: [str] A two letter country code

    :return: dict

    Usage::

            from pygbif import occurrences
            occurrences.count_countries(publishingCountry = "DE")
    """
    url = gbif_baseurl + "occurrence/counts/countries"
    out = gbif_GET(url, {"publishingCountry": publishingCountry}, **kwargs)
    return out


def count_publishingcountries(country, **kwargs):
    """
    Lists occurrence counts for all countries that publish data about the given country

    :param country: [str] A country, two letter code

    :return: dict

    Usage::

            from pygbif import occurrences
            occurrences.count_publishingcountries(country = "DE")
    """
    url = gbif_baseurl + "occurrence/counts/publishingCountries"
    out = gbif_GET(url, {"country": country}, **kwargs)
    return out


def count_schema(**kwargs):
    """
    List the supported metrics by the service

    :return: dict

    Usage::

            from pygbif import occurrences
            occurrences.count_schema()
    """
    url = gbif_baseurl + "occurrence/count/schema"
    out = gbif_GET(url, {}, **kwargs)
    return out



"""tests for institution search"""
import vcr
from pygbif import institution

keyz = ["count", "results", "endOfRecords", "limit", "offset"]

@vcr.use_cassette("test/vcr_cassettes/test_institution_search.yaml")
def test_search():
    "institution.search - basic test"
    res = institution.search(limit=10)
    assert "dict" == res.__class__.__name__
    assert 5 == len(res)
    assert sorted(keyz) == sorted(res.keys())
    assert 10 == len(res["results"])
    assert res["count"] >= 8000

@vcr.use_cassette("test/vcr_cassettes/test_institution_search_q.yaml")
def test_search_q():
    "institution.search - q"
    res = institution.search(q="Kansas", limit=10)
    assert "dict" == res.__class__.__name__
    assert 5 == len(res)
    assert sorted(keyz) == sorted(res.keys())
    assert 10 == len(res["results"])
    assert res["count"] >= 10
    assert res["count"] <= 8000

@vcr.use_cassette("test/vcr_cassettes/test_institution_search_country.yaml")
def test_search_country():
    "institution.search - country"
    res = institution.search(country=["US", "GB"], limit=10)
    assert "dict" == res.__class__.__name__
    assert 5 == len(res)
    assert sorted(keyz) == sorted(res.keys())
    assert 10 == len(res["results"])
    assert res["count"] >= 1000
    assert res["count"] <= 8000

@vcr.use_cassette("test/vcr_cassettes/test_institution_search_active.yaml")
def test_search_active():
    "institution.search - active"
    res = institution.search(active=True, limit=10)
    assert "dict" == res.__class__.__name__
    assert 5 == len(res)
    assert sorted(keyz) == sorted(res.keys())
    assert 10 == len(res["results"])
    for r in res["results"]:
        assert r["active"] is True

@vcr.use_cassette("test/vcr_cassettes/test_institution_typeSpecimenCount.yaml")
def test_search_typeSpecimenCount():
    "institution.search - typeSpecimenCount"
    res = institution.search(typeSpecimenCount="10,100", limit=10)
    assert "dict" == res.__class__.__name__
    assert 5 == len(res)
    assert sorted(keyz) == sorted(res.keys())
    assert 10 == len(res["results"])
    for r in res["results"]:
        assert r["typeSpecimenCount"] >= 10
        assert r["typeSpecimenCount"] <= 100




import re

from pygbif.gbifutils import (
    gbif_baseurl,
    gbif_GET,
    len2,
    bool2str,
    check_data,
    requests_argset,
)


def dataset_metrics(uuid):
    """
    Get details on a GBIF dataset.

    :param uuid: [str] One or more dataset UUIDs. See examples.

    References: http://www.gbif.org/developer/registry#datasetMetrics

    Usage::

            from pygbif import registry
            registry.dataset_metrics(uuid='3f8a1297-3259-4700-91fc-acc4170b27ce')
            registry.dataset_metrics(uuid='66dd0960-2d7d-46ee-a491-87b9adcfe7b1')
            registry.dataset_metrics(uuid=['3f8a1297-3259-4700-91fc-acc4170b27ce', '66dd0960-2d7d-46ee-a491-87b9adcfe7b1'])
    """

    def getdata(x):
        url = gbif_baseurl + "dataset/" + x + "/metrics"
        return gbif_GET(url, {})

    if len2(uuid) == 1:
        return getdata(uuid)
    else:
        return [getdata(x) for x in uuid]


def datasets(
    data="all",
    type=None,
    uuid=None,
    query=None,
    id=None,
    limit=100,
    offset=None,
    **kwargs
):
    """
    Search for datasets and dataset metadata.

    :param data: [str] The type of data to get. Default: ``all``
    :param type: [str] Type of dataset, options include ``OCCURRENCE``, etc.
    :param uuid: [str] UUID of the data node provider. This must be specified if data
         is anything other than ``all``.
    :param query: [str] Query term(s). Only used when ``data = 'all'``
    :param id: [int] A metadata document id.

    References http://www.gbif.org/developer/registry#datasets

    Usage::

            from pygbif import registry
            registry.datasets(limit=5)
            registry.datasets(type="OCCURRENCE")
            registry.datasets(uuid="a6998220-7e3a-485d-9cd6-73076bd85657")
            registry.datasets(data='contact', uuid="a6998220-7e3a-485d-9cd6-73076bd85657")
            registry.datasets(data='metadata', uuid="a6998220-7e3a-485d-9cd6-73076bd85657")
            registry.datasets(data='metadata', uuid="a6998220-7e3a-485d-9cd6-73076bd85657", id=598)
            registry.datasets(data=['deleted','duplicate'])
            registry.datasets(data=['deleted','duplicate'], limit=1)
    """
    args = {"q": query, "type": type, "limit": limit, "offset": offset}
    data_choices = [
        "all",
        "organization",
        "contact",
        "endpoint",
        "identifier",
        "tag",
        "machinetag",
        "comment",
        "constituents",
        "document",
        "metadata",
        "deleted",
        "duplicate",
        "subDataset",
        "withNoEndpoint",
    ]
    check_data(data, data_choices)
    if len2(data) == 1:
        return datasets_fetch(data, uuid, args, **kwargs)
    else:
        return [datasets_fetch(x, uuid, args, **kwargs) for x in data]


def datasets_fetch(x, uuid, args, **kwargs):
    if (
        x not in ["all", "deleted", "duplicate", "subDataset", "withNoEndpoint"]
        and uuid is None
    ):
        raise TypeError(
            "You must specify a uuid if data does not equal all and data does not equal of deleted, duplicate, subDataset, or withNoEndpoint"
        )

    if uuid is None:
        if x == "all":
            url = gbif_baseurl + "dataset"
        else:
            if id is not None and x == "metadata":
                url = gbif_baseurl + "dataset/metadata/" + id + "/document"
            else:
                url = gbif_baseurl + "dataset/" + x
    else:
        if x == "all":
            url = gbif_baseurl + "dataset/" + uuid
        else:
            url = gbif_baseurl + "dataset/" + uuid + "/" + x

    res = gbif_GET(url, args, **kwargs)
    return res


def dataset_suggest(
    q=None,
    type=None,
    keyword=None,
    owningOrg=None,
    publishingOrg=None,
    hostingOrg=None,
    publishingCountry=None,
    decade=None,
    limit=100,
    offset=None,
    **kwargs
):
    """
    Search that returns up to 20 matching datasets. Results are ordered by relevance.


    :param q: [str] Query term(s) for full text search.  The value for this parameter can be a simple word or a phrase. Wildcards can be added to the simple word parameters only, e.g. ``q=*puma*``
    :param type: [str] Type of dataset, options include OCCURRENCE, etc.
    :param keyword: [str] Keyword to search by. Datasets can be tagged by keywords, which you can search on. The search is done on the merged collection of tags, the dataset keywordCollections and temporalCoverages. SEEMS TO NOT BE WORKING ANYMORE AS OF 2016-09-02.
    :param owningOrg: [str] Owning organization. A uuid string. See :func:`~pygbif.registry.organizations`
    :param publishingOrg: [str] Publishing organization. A uuid string. See :func:`~pygbif.registry.organizations`
    :param hostingOrg: [str] Hosting organization. A uuid string. See :func:`~pygbif.registry.organizations`
    :param publishingCountry: [str] Publishing country.
    :param decade: [str] Decade, e.g., 1980. Filters datasets by their temporal coverage broken down to decades. Decades are given as a full year, e.g. 1880, 1960, 2000, etc, and will return datasets wholly contained in the decade as well as those that cover the entire decade or more. Facet by decade to get the break down, e.g. ``/search?facet=DECADE&facet_only=true`` (see example below)
    :param limit: [int] Number of results to return. Default: ``300``
    :param offset: [int] Record to start at. Default: ``0``

    :return: A dictionary

    References: http://www.gbif.org/developer/registry#datasetSearch

    Usage::

            from pygbif import registry
            registry.dataset_suggest(q="Amazon", type="OCCURRENCE")

            # Suggest datasets tagged with keyword "france".
            registry.dataset_suggest(keyword="france")

            # Suggest datasets owned by the organization with key
            # "07f617d0-c688-11d8-bf62-b8a03c50a862" (UK NBN).
            registry.dataset_suggest(owningOrg="07f617d0-c688-11d8-bf62-b8a03c50a862")

            # Fulltext search for all datasets having the word "amsterdam" somewhere in
            # its metadata (title, description, etc).
            registry.dataset_suggest(q="amsterdam")

            # Limited search
            registry.dataset_suggest(type="OCCURRENCE", limit=2)
            registry.dataset_suggest(type="OCCURRENCE", limit=2, offset=10)

            # Return just descriptions
            registry.dataset_suggest(type="OCCURRENCE", limit = 5, description=True)

            # Search by decade
            registry.dataset_suggest(decade=1980, limit = 30)
    """
    url = gbif_baseurl + "dataset/suggest"
    args = {
        "q": q,
        "type": type,
        "keyword": keyword,
        "publishingOrg": publishingOrg,
        "hostingOrg": hostingOrg,
        "owningOrg": owningOrg,
        "decade": decade,
        "publishingCountry": publishingCountry,
        "limit": limit,
        "offset": offset,
    }
    out = gbif_GET(url, args, **kwargs)
    return out


def dataset_search(
    q=None,
    type=None,
    keyword=None,
    owningOrg=None,
    publishingOrg=None,
    hostingOrg=None,
    decade=None,
    publishingCountry=None,
    facet=None,
    facetMincount=None,
    facetMultiselect=None,
    hl=False,
    limit=100,
    offset=None,
    **kwargs
):
    """
    Full text search across all datasets. Results are ordered by relevance.

    :param q: [str] Query term(s) for full text search.  The value for this parameter
         can be a simple word or a phrase. Wildcards can be added to the simple word
         parameters only, e.g. ``q=*puma*``
    :param type: [str] Type of dataset, options include OCCURRENCE, etc.
    :param keyword: [str] Keyword to search by. Datasets can be tagged by keywords, which
         you can search on. The search is done on the merged collection of tags, the
         dataset keywordCollections and temporalCoverages. SEEMS TO NOT BE WORKING
         ANYMORE AS OF 2016-09-02.
    :param owningOrg: [str] Owning organization. A uuid string. See :func:`~pygbif.registry.organizations`
    :param publishingOrg: [str] Publishing organization. A uuid string. See :func:`~pygbif.registry.organizations`
    :param hostingOrg: [str] Hosting organization. A uuid string. See :func:`~pygbif.registry.organizations`
    :param publishingCountry: [str] Publishing country.
    :param decade: [str] Decade, e.g., 1980. Filters datasets by their temporal coverage
         broken down to decades. Decades are given as a full year, e.g. 1880, 1960, 2000,
         etc, and will return datasets wholly contained in the decade as well as those
         that cover the entire decade or more. Facet by decade to get the break down,
         e.g. ``/search?facet=DECADE&facet_only=true`` (see example below)
    :param facet: [str] A list of facet names used to retrieve the 100 most frequent values
            for a field. Allowed facets are: type, keyword, publishingOrg, hostingOrg, decade,
            and publishingCountry. Additionally subtype and country are legal values but not
            yet implemented, so data will not yet be returned for them.
    :param facetMincount: [str] Used in combination with the facet parameter. Set
            facetMincount={#} to exclude facets with a count less than {#}, e.g.
            http://api.gbif.org/v1/dataset/search?facet=type&limit=0&facetMincount=10000
            only shows the type value 'OCCURRENCE' because 'CHECKLIST' and 'METADATA' have
            counts less than 10000.
    :param facetMultiselect: [bool] Used in combination with the facet parameter. Set
            ``facetMultiselect=True`` to still return counts for values that are not currently
            filtered, e.g.
            http://api.gbif.org/v1/dataset/search?facet=type&limit=0&type=CHECKLIST&facetMultiselect=true
            still shows type values 'OCCURRENCE' and 'METADATA' even though type is being
            filtered by ``type=CHECKLIST``
    :param hl: [bool] Set ``hl=True`` to highlight terms matching the query when in fulltext
            search fields. The highlight will be an emphasis tag of class 'gbifH1' e.g.
            http://api.gbif.org/v1/dataset/search?q=plant&hl=true
            Fulltext search fields include: title, keyword, country, publishing country,
            publishing organization title, hosting organization title, and description. One
            additional full text field is searched which includes information from metadata
            documents, but the text of this field is not returned in the response.
    :param limit: [int] Number of results to return. Default: ``300``
    :param offset: [int] Record to start at. Default: ``0``

    :note: Note that you can pass in additional faceting parameters on a per field basis.
            For example, if you want to limit the numbef of facets returned from a field ``foo`` to
            3 results, pass in ``foo_facetLimit = 3``. GBIF does not allow all per field parameters,
            but does allow some. See also examples.

    :return: A dictionary

    References: http://www.gbif.org/developer/registry#datasetSearch

    Usage::

            from pygbif import registry
            # Gets all datasets of type "OCCURRENCE".
            registry.dataset_search(type="OCCURRENCE", limit = 10)

            # Fulltext search for all datasets having the word "amsterdam" somewhere in
            # its metadata (title, description, etc).
            registry.dataset_search(q="amsterdam", limit = 10)

            # Limited search
            registry.dataset_search(type="OCCURRENCE", limit=2)
            registry.dataset_search(type="OCCURRENCE", limit=2, offset=10)

            # Search by decade
            registry.dataset_search(decade=1980, limit = 10)

            # Faceting
            ## just facets
            registry.dataset_search(facet="decade", facetMincount=10, limit=0)

            ## data and facets
            registry.dataset_search(facet="decade", facetMincount=10, limit=2)

            ## many facet variables
            registry.dataset_search(facet=["decade", "type"], facetMincount=10, limit=0)

            ## facet vars
            ### per variable paging
            x = registry.dataset_search(
                facet = ["decade", "type"],
                decade_facetLimit = 3,
                type_facetLimit = 3,
                limit = 0
            )

            ## highlight
            x = registry.dataset_search(q="plant", hl=True, limit = 10)
            [ z['description'] for z in x['results'] ]
    """
    url = gbif_baseurl + "dataset/search"
    args = {
        "q": q,
        "type": type,
        "keyword": keyword,
        "owningOrg": owningOrg,
        "publishingOrg": publishingOrg,
        "hostingOrg": hostingOrg,
        "decade": decade,
        "publishingCountry": publishingCountry,
        "facet": facet,
        "facetMincount": facetMincount,
        "facetMultiselect": bool2str(facetMultiselect),
        "hl": bool2str(hl),
        "limit": limit,
        "offset": offset,
    }
    gbif_kwargs = {key: kwargs[key] for key in kwargs if key not in requests_argset}
    if gbif_kwargs is not None:
        xx = dict(
            zip([re.sub("_", ".", x) for x in gbif_kwargs.keys()], gbif_kwargs.values())
        )
        args.update(xx)
    kwargs = {key: kwargs[key] for key in kwargs if key in requests_argset}
    out = gbif_GET(url, args, **kwargs)
    return out



"""Tests for occurrences module - search methods"""
import vcr
from pygbif import occurrences

keyz = ["count", "facets", "results", "endOfRecords", "limit", "offset"]
x = "https://orcid.org/0000-0003-1691-239X"


@vcr.use_cassette("test/vcr_cassettes/test_search.yaml")
def test_search():
    "occurrences.search - basic test"
    res = occurrences.search(taxonKey=3329049)
    assert "dict" == res.__class__.__name__
    assert 6 == len(res)
    assert sorted(keyz) == sorted(res.keys())


@vcr.use_cassette("test/vcr_cassettes/test_search_key1.yaml")
def test_search_key1():
    "occurrences.search - diff taxonKey"
    res = occurrences.search(taxonKey=2431762)
    assert "dict" == res.__class__.__name__
    assert 6 == len(res)
    assert 2431762 == res["results"][0]["taxonKey"]


@vcr.use_cassette("test/vcr_cassettes/test_search_key2.yaml")
def test_search_key2():
    "occurrences.search - diff taxonKey2"
    res = occurrences.search(taxonKey=2683264)
    assert "dict" == res.__class__.__name__
    assert 6 == len(res)
    assert 2683264 == res["results"][0]["taxonKey"]


@vcr.use_cassette("test/vcr_cassettes/test_search_recorded_by_id.yaml")
def test_search_recorded_by_id():
    "occurrences.search - recordedByID"
    res = occurrences.search(recordedByID=x, limit=3)
    assert "dict" == res.__class__.__name__
    assert 6 == len(res)
    assert x == res["results"][0]["recordedByIDs"][0]["value"]


@vcr.use_cassette("test/vcr_cassettes/test_search_identified_by_id.yaml")
def test_search_identified_by_id():
    "occurrences.search - identifiedByID"
    res = occurrences.search(identifiedByID=x, limit=3)
    assert "dict" == res.__class__.__name__
    assert 6 == len(res)
    assert x == res["results"][0]["identifiedByIDs"][0]["value"]



from geojson_rewind import rewind
from geomet import wkt
import decimal
import statistics


def wkt_rewind(x, digits=None):
    """
    reverse WKT winding order

    :param x: [str] WKT string
    :param digits: [int] number of digits after decimal to use for the return string. 
        by default, we use the mean number of digits in your string.

    :return: a string

    Usage::
        
        from pygbif import wkt_rewind
        x = 'POLYGON((144.6 13.2, 144.6 13.6, 144.9 13.6, 144.9 13.2, 144.6 13.2))'
        wkt_rewind(x)
        wkt_rewind(x, digits = 0)
        wkt_rewind(x, digits = 3)
        wkt_rewind(x, digits = 7)
    """
    z = wkt.loads(x)
    if digits is None:
        coords = z["coordinates"]
        nums = __flatten(coords)
        dec_n = [decimal.Decimal(str(w)).as_tuple().exponent for w in nums]
        digits = abs(statistics.mean(dec_n))
    else:
        if not isinstance(digits, int):
            raise TypeError("'digits' must be an int")
    wound = rewind(z)
    back_to_wkt = wkt.dumps(wound, decimals=digits)
    return back_to_wkt


# from https://stackoverflow.com/a/12472564/1091766
def __flatten(S):
    if S == []:
        return S
    if isinstance(S[0], list):
        return __flatten(S[0]) + __flatten(S[1:])
    return S[:1] + __flatten(S[1:])



# pygbif

"""
pygbif library
~~~~~~~~~~~~~~~~~~~~~

pygbif is a Python client for the Global Biodiversity Information Facility (GBIF) API.

Usage::

        # Import entire library
        import pygbif
        # or import modules as needed
        ## occurrences
        from pygbif import occurrences
        ## species
        from pygbif import species
        ## registry
        from pygbif import registry

        ## use advanced logging
        ### setup first
        import requests
        import logging
        import http.client
        http.client.HTTPConnection.debuglevel = 1
        logging.basicConfig()
        logging.getLogger().setLevel(logging.DEBUG)
        requests_log = logging.getLogger("requests.packages.urllib3")
        requests_log.setLevel(logging.DEBUG)
        requests_log.propagate = True
        ### then make request
        from pygbif import occurrences
        occurrences.search(geometry='POLYGON((30.1 10.1, 10 20, 20 40, 40 40, 30.1 10.1))', limit=20)
"""

from .package_metadata import __author__, __license__, __version__, __title__
from .occurrences import search, get, count, download
from .species import name_parser, name_suggest, name_backbone, name_lookup, name_usage
from .registry import datasets, nodes, networks, organizations, installations
from .maps import map, GbifMap
from .gbifissues import occ_issues_lookup
from .utils import *
from .caching import caching
from .literature import search 
from .collection import search
from .institution import search

# Set default logging handler to avoid "No handler found" warnings.
import logging
from logging import NullHandler

logging.getLogger(__name__).addHandler(NullHandler())
logging.basicConfig(format="%(levelname)s:%(message)s", level=logging.INFO)



"""Tests for maps module - maps"""
import pytest
import unittest
import vcr
import requests
import matplotlib

matplotlib.use("Agg")
import pygbif


class TestMapsClass(unittest.TestCase):
    @vcr.use_cassette("test/vcr_cassettes/test_map.yaml")
    def test_map(self):
        "maps.map - basic test"
        res = pygbif.maps.map(taxonKey=2435098)
        self.assertIsInstance(res, pygbif.maps.GbifMap)
        self.assertIsInstance(res.response, requests.Response)
        self.assertIsInstance(res.path, str)
        self.assertIsInstance(res.img, matplotlib.image.AxesImage)

    def test_map_year_range(self):
        "maps.map - year range"
        res = pygbif.maps.map(taxonKey=2435098, year=range(2007, 2011 + 1))
        self.assertIsInstance(res, pygbif.maps.GbifMap)
        self.assertRegex(res.response.request.path_url, "2007%2C2011")
        # self.assertIsInstance(res.path, str)
        # self.assertIsInstance(res.img, matplotlib.image.AxesImage)

    def test_map_basisofrecord_str_class(self):
        "maps.map - basisofrecord"
        res = pygbif.maps.map(
            taxonKey=2480498, year=2010, basisOfRecord="HUMAN_OBSERVATION"
        )
        self.assertIsInstance(res, pygbif.maps.GbifMap)
        self.assertRegex(res.response.request.path_url, "basisOfRecord")
        self.assertRegex(res.response.request.path_url, "HUMAN_OBSERVATION")

    def test_map_basisofrecord_list_class(self):
        "maps.map - basisofrecord"
        res = pygbif.maps.map(
            taxonKey=2480498,
            year=2010,
            basisOfRecord=["HUMAN_OBSERVATION", "LIVING_SPECIMEN"],
        )
        self.assertIsInstance(res, pygbif.maps.GbifMap)
        self.assertRegex(res.response.request.path_url, "basisOfRecord")
        self.assertRegex(res.response.request.path_url, "HUMAN_OBSERVATION")
        self.assertRegex(res.response.request.path_url, "LIVING_SPECIMEN")

    def test_maps_fails_well(self):
        "maps.map - fails well"
        with pytest.raises(ValueError):
            pygbif.maps.map(year=2300)
            pygbif.maps.map(year="2010")
            pygbif.maps.map(basisOfRecord="foobar")
            pygbif.maps.map(format="foobar")
            pygbif.maps.map(source="foobar")
            pygbif.maps.map(srs="foobar")
            pygbif.maps.map(bin="foobar")
            pygbif.maps.map(style="foobar")



import re

from pygbif.gbifutils import bool2str, bn, requests_argset, gbif_GET, gbif_baseurl


def name_lookup(
    q=None,
    rank=None,
    higherTaxonKey=None,
    status=None,
    isExtinct=None,
    habitat=None,
    nameType=None,
    datasetKey=None,
    nomenclaturalStatus=None,
    limit=100,
    offset=None,
    facet=False,
    facetMincount=None,
    facetMultiselect=None,
    type=None,
    hl=False,
    verbose=False,
    **kwargs
):
    """
	Lookup names in all taxonomies in GBIF.

	This service uses fuzzy lookup so that you can put in partial names and
	you should get back those things that match. See examples below.

	:param q: [str] Query term(s) for full text search (optional)
	:param rank: [str] ``CLASS``, ``CULTIVAR``, ``CULTIVAR_GROUP``, ``DOMAIN``, ``FAMILY``,
		 ``FORM``, ``GENUS``, ``INFORMAL``, ``INFRAGENERIC_NAME``, ``INFRAORDER``, ``INFRASPECIFIC_NAME``,
		 ``INFRASUBSPECIFIC_NAME``, ``KINGDOM``, ``ORDER``, ``PHYLUM``, ``SECTION``, ``SERIES``, ``SPECIES``, ``STRAIN``, ``SUBCLASS``,
		 ``SUBFAMILY``, ``SUBFORM``, ``SUBGENUS``, ``SUBKINGDOM``, ``SUBORDER``, ``SUBPHYLUM``, ``SUBSECTION``, ``SUBSERIES``,
		 ``SUBSPECIES``, ``SUBTRIBE``, ``SUBVARIETY``, ``SUPERCLASS``, ``SUPERFAMILY``, ``SUPERORDER``, ``SUPERPHYLUM``,
		 ``SUPRAGENERIC_NAME``, ``TRIBE``, ``UNRANKED``, ``VARIETY`` (optional)
	:param verbose: [bool] If ``True`` show alternative matches considered which had been rejected.
	:param higherTaxonKey: [str] Filters by any of the higher Linnean rank keys. Note this
			is within the respective checklist and not searching nub keys across all checklists (optional)
	:param status: [str] (optional) Filters by the taxonomic status as one of:

		* ``ACCEPTED``
		* ``DETERMINATION_SYNONYM`` Used for unknown child taxa referred to via spec, ssp, ...
		* ``DOUBTFUL`` Treated as accepted, but doubtful whether this is correct.
		* ``HETEROTYPIC_SYNONYM`` More specific subclass of ``SYNONYM``.
		* ``HOMOTYPIC_SYNONYM`` More specific subclass of ``SYNONYM``.
		* ``INTERMEDIATE_RANK_SYNONYM`` Used in nub only.
		* ``MISAPPLIED`` More specific subclass of ``SYNONYM``.
		* ``PROPARTE_SYNONYM`` More specific subclass of ``SYNONYM``.
		* ``SYNONYM`` A general synonym, the exact type is unknown.

	:param isExtinct: [bool] Filters by extinction status (e.g. ``isExtinct=True``)
	:param habitat: [str] Filters by habitat. One of: ``marine``, ``freshwater``, or
			``terrestrial`` (optional)
	:param nameType: [str] (optional) Filters by the name type as one of:

		* ``BLACKLISTED`` surely not a scientific name.
		* ``CANDIDATUS`` Candidatus is a component of the taxonomic name for a bacterium that cannot be maintained in a Bacteriology Culture Collection.
		* ``CULTIVAR`` a cultivated plant name.
		* ``DOUBTFUL`` doubtful whether this is a scientific name at all.
		* ``HYBRID`` a hybrid formula (not a hybrid name).
		* ``INFORMAL`` a scientific name with some informal addition like "cf." or indetermined like Abies spec.
		* ``SCINAME`` a scientific name which is not well formed.
		* ``VIRUS`` a virus name.
		* ``WELLFORMED`` a well formed scientific name according to present nomenclatural rules.

	:param datasetKey: [str] Filters by the dataset's key (a uuid) (optional)
	:param nomenclaturalStatus: [str] Not yet implemented, but will eventually allow for
			filtering by a nomenclatural status enum
	:param limit: [fixnum] Number of records to return. Maximum: ``1000``. (optional)
	:param offset: [fixnum] Record number to start at. (optional)
	:param facet: [str] A list of facet names used to retrieve the 100 most frequent values
			for a field. Allowed facets are: ``datasetKey``, ``higherTaxonKey``, ``rank``, ``status``,
			``isExtinct``, ``habitat``, and ``nameType``. Additionally ``threat`` and ``nomenclaturalStatus``
			are legal values but not yet implemented, so data will not yet be returned for them. (optional)
	:param facetMincount: [str] Used in combination with the facet parameter. Set
			``facetMincount={#}`` to exclude facets with a count less than {#}, e.g.
			http://bit.ly/1bMdByP only shows the type value ``ACCEPTED`` because the other
			statuses have counts less than 7,000,000 (optional)
	:param facetMultiselect: [bool] Used in combination with the facet parameter. Set
			``facetMultiselect=True`` to still return counts for values that are not currently
			filtered, e.g. http://bit.ly/19YLXPO still shows all status values even though
			status is being filtered by ``status=ACCEPTED`` (optional)
	:param type: [str] Type of name. One of ``occurrence``, ``checklist``, or ``metadata``. (optional)
	:param hl: [bool] Set ``hl=True`` to highlight terms matching the query when in fulltext
			search fields. The highlight will be an emphasis tag of class ``gbifH1`` e.g.
			``q='plant', hl=True``. Fulltext search fields include: ``title``, ``keyword``, ``country``,
			``publishing country``, ``publishing organization title``, ``hosting organization title``, and
			``description``. One additional full text field is searched which includes information from
			metadata documents, but the text of this field is not returned in the response. (optional)

	:return: A dictionary

	:references: http://www.gbif.org/developer/species#searching

	Usage::

			from pygbif import species

			# Look up names like mammalia
			species.name_lookup(q='mammalia')

			# Paging
			species.name_lookup(q='mammalia', limit=1)
			species.name_lookup(q='mammalia', limit=1, offset=2)

			# large requests, use offset parameter
			first = species.name_lookup(q='mammalia', limit=1000)
			second = species.name_lookup(q='mammalia', limit=1000, offset=1000)

			# Get all data and parse it, removing descriptions which can be quite long
			species.name_lookup('Helianthus annuus', rank="species", verbose=True)

			# Get all data and parse it, removing descriptions field which can be quite long
			out = species.name_lookup('Helianthus annuus', rank="species")
			res = out['results']
			[ z.pop('descriptions', None) for z in res ]
			res

			# Fuzzy searching
			species.name_lookup(q='Heli', rank="genus")

			# Limit records to certain number
			species.name_lookup('Helianthus annuus', rank="species", limit=2)

			# Query by habitat
			species.name_lookup(habitat = "terrestrial", limit=2)
			species.name_lookup(habitat = "marine", limit=2)
			species.name_lookup(habitat = "freshwater", limit=2)

			# Using faceting
			species.name_lookup(facet='status', limit=0, facetMincount='70000')
			species.name_lookup(facet=['status', 'higherTaxonKey'], limit=0, facetMincount='700000')

			species.name_lookup(facet='nameType', limit=0)
			species.name_lookup(facet='habitat', limit=0)
			species.name_lookup(facet='datasetKey', limit=0)
			species.name_lookup(facet='rank', limit=0)
			species.name_lookup(facet='isExtinct', limit=0)

			# text highlighting
			species.name_lookup(q='plant', hl=True, limit=30)

			# Lookup by datasetKey
			species.name_lookup(datasetKey='3f8a1297-3259-4700-91fc-acc4170b27ce')
	"""
    args = {
        "q": q,
        "rank": rank,
        "higherTaxonKey": higherTaxonKey,
        "status": status,
        "isExtinct": bool2str(isExtinct),
        "habitat": habitat,
        "nameType": nameType,
        "datasetKey": datasetKey,
        "nomenclaturalStatus": nomenclaturalStatus,
        "limit": limit,
        "offset": offset,
        "facet": bn(facet),
        "facetMincount": facetMincount,
        "facetMultiselect": bool2str(facetMultiselect),
        "hl": bool2str(hl),
        "verbose": bool2str(verbose),
        "type": type,
    }
    gbif_kwargs = {key: kwargs[key] for key in kwargs if key not in requests_argset}
    if gbif_kwargs is not None:
        xx = dict(
            zip([re.sub("_", ".", x) for x in gbif_kwargs.keys()], gbif_kwargs.values())
        )
        args.update(xx)
    kwargs = {key: kwargs[key] for key in kwargs if key in requests_argset}
    return gbif_GET(gbif_baseurl + "species/search", args, **kwargs)



"""tests for collection search"""
import vcr
from pygbif import collection

keyz = ["count", "results", "endOfRecords", "limit", "offset"]

@vcr.use_cassette("test/vcr_cassettes/test_collection_search.yaml")
def test_search():
    "collection.search - basic test"
    res = collection.search(limit=10)
    assert "dict" == res.__class__.__name__
    assert 5 == len(res)
    assert sorted(keyz) == sorted(res.keys())
    assert 10 == len(res["results"])
    assert res["count"] >= 10000

@vcr.use_cassette("test/vcr_cassettes/test_collection_search_q.yaml")
def test_search_q():
    "collection.search - q"
    res = collection.search(q="Kansas", limit=10)
    assert "dict" == res.__class__.__name__
    assert 5 == len(res)
    assert sorted(keyz) == sorted(res.keys())
    assert 10 == len(res["results"])
    assert res["count"] >= 30
    assert res["count"] <= 10000

@vcr.use_cassette("test/vcr_cassettes/test_collection_search_country.yaml")
def test_search_country():
    "collection.search - country"
    res = collection.search(country=["US","GB"], limit=10)
    assert "dict" == res.__class__.__name__
    assert 5 == len(res)
    assert sorted(keyz) == sorted(res.keys())
    assert 10 == len(res["results"])
    assert res["count"] >= 3000
    assert res["count"] <= 10000

@vcr.use_cassette("test/vcr_cassettes/test_collection_search_active.yaml")
def test_search_active():
    "collection.search - active"
    res = collection.search(active=True, limit=10)
    assert "dict" == res.__class__.__name__
    assert 5 == len(res)
    assert sorted(keyz) == sorted(res.keys())
    assert 10 == len(res["results"])
    assert res["count"] >= 9000
    for r in res["results"]:
        assert r["active"] is True

@vcr.use_cassette("test/vcr_cassettes/test_collection_search_institutionKey.yaml")
def test_search_institutionKey():
    "collection.search - institutionKey"
    res = collection.search(institutionKey="6a6ac6c5-1b8a-48db-91a2-f8661274ff80", limit=10)
    assert "dict" == res.__class__.__name__
    assert 5 == len(res)
    assert sorted(keyz) == sorted(res.keys())
    assert 10 == len(res["results"])
    assert res["count"] >= 20
    for r in res["results"]:
        assert r["institutionKey"] == "6a6ac6c5-1b8a-48db-91a2-f8661274ff80"




from pygbif.gbifutils import gbif_baseurl, bool2str, requests_argset, gbif_GET

def search(
    citationType=None,
    countriesOfCoverage=None,
    countriesOfResearcher=None,
    doi=None,
    gbifDatasetKey=None,
    gbifDownloadKey=None,
    gbifHigherTaxonKey=None,
    gbifNetworkKey=None,
    gbifOccurrenceKey=None,
    gbifProjectIdentifier=None,
    gbifProgrammeAcronym=None,
    gbifTaxonKey=None,
    literatureType=None,
    openAccess=None,
    peerReview=None,
    publisher=None,
    publishingOrganizationKey=None,
    relevance=None,
    source=None,
    topics=None,
    websites=None,
    year=None,
    language=None,
    added=None,
    published=None,
    discovered=None,
    modified=None,
    hl=None,
    limit=100,
    offset=0,
    facet=None,
    facetMincount=None,
    facetMultiselect=None,
    facetLimit=None,
    facetOffset=None,
    q=None,
    **kwargs
):
    """
    Search for literature indexed by GBIF
    
    :param citationType: [str] The manner in which GBIF is cited in a paper.
    :param countriesOfCoverage: [str], Countries of coverage
    :param countriesOfResearcher: [str] Country or area of institution with which author is affiliated. 
    :param doi: [str] Digital Object Identifier (DOI) of the literature item.
    :param gbifDatasetKey: [str] GBIF dataset key uuid referenced in publication.
    :param gbifDownloadKey: [str] GBIF download referenced in publication.
    :param gbifHigherTaxonKey: [str] All parent keys of any taxon that is the focus of the paper.
    :param gbifNetworkKey: [str] GBIF network referenced in publication.
    :param gbifOccurrenceKey: [str] Any GBIF occurrence keys directly mentioned in a paper.
    :param gbifProjectIdentifier: [str] GBIF dataset referenced in publication.
    :param gbifProgrammeAcronym: [str] GBIF dataset referenced in publication.
    :param gbifTaxonKey: [str] Key(s) from the GBIF backbone of taxa that are the focus of a paper.
    :param literatureType: [str] Type of literature, e.g. journal article.
    :param openAccess: [bool] Is the publication Open Access?
    :param peerReview: [bool] Has the publication undergone peer review?
    :param publisher: [str] Publisher of journal.
    :param publishingOrganizationKey: [str] Publisher whose dataset is referenced in publication.
    :param relevance: [str] Relevance to GBIF community.
    :param source: [str] Journal of publication.
    :param topics: [str] Topic of publication.
    :param websites: [str] Website of publication.
    :param year: [str] Year of publication. This can be a single range such as "2019,2021".
    :param language: [str] 3 letter language code of publication. e.g. "eng"
    :param added: [str] Date or date range when the publication was added. Format is ISO 8601, e.g., '2024-07-14' or '2024-07-14,2024-08-14'.
    :param published: [str] Date or date range when the publication was published. Format is ISO 8601, e.g., '2024-02-22' or '2024-02-22,2024-03-22'.
    :param discovered: [str] Date or date range when the publication was discovered. Format is ISO 8601, e.g., '2024-02-26' or '2024-02-26,2024-03-26'.
    :param modified: [str] Date or date range when the publication was discovered. Format is ISO 8601, e.g., '2024-07-26' or '2024-07-26,2024-10-26'.
    :param hl: [str] Set hl=true to highlight terms matching the query when in fulltext search fields. 
    :param limit: [int] Controls the number of results in the page.
    :param offset: [int] Determines the offset for the search results.
    :param facet: [str] A facet name used to retrieve the most frequent values for a field.
    :param facetMincount: [int] Used in combination with the facet parameter. Set facetMincount={#} to exclude facets with a count less than {#}.
    :param facetMultiselect: [bool] Used in combination with the facet parameter. Set facetMultiselect=true to still return counts for values that are not currently filtered.
    :param facetLimit: [int] Limit the number of facet values returned.
    :param facetOffset: [int] Facet parameters allow paging requests using the parameters facetOffset and facetLimit.
    :param q: [str] Full text query. 
    
    :return: A dictionary

    Usage::

        from pygbif import literature as lit
        lit.search(limit=10) # basic search for 10 records

        # search for a specific download key
        lit.search(gbifDownloadKey="0235283-220831081235567") 

        # search for a year range
        lit.search(year="2010,2024", limit=10)

        # search for count only 
        lit.search(year="2010,2024", limit=0)["count"]    
    """
    url = gbif_baseurl + "literature/search"
    args = {
        "citationType": citationType,
        "countriesOfCoverage": countriesOfCoverage,
        "countriesOfResearcher": countriesOfResearcher,
        "doi": doi,
        "gbifDatasetKey": gbifDatasetKey,
        "gbifDownloadKey": gbifDownloadKey,
        "gbifHigherTaxonKey": gbifHigherTaxonKey,
        "gbifNetworkKey": gbifNetworkKey,
        "gbifOccurrenceKey": gbifOccurrenceKey,
        "gbifProjectIdentifier": gbifProjectIdentifier,
        "gbifProgrammeAcronym": gbifProgrammeAcronym,
        "gbifTaxonKey": gbifTaxonKey,
        "literatureType": literatureType,
        "openAccess": openAccess,
        "peerReview": peerReview,
        "publisher": publisher,
        "publishingOrganizationKey": publishingOrganizationKey,
        "relevance": relevance,
        "source": source,
        "topics": topics,
        "websites": websites,
        "year": year,
        "language": language,
        "added": added,
        "published": published,
        "discovered": discovered,
        "modified": modified,
        "hl": hl,
        "limit": limit,
        "offset": offset,
        "facet": facet,
        "facetMincount": facetMincount,
        "facetMultiselect": facetMultiselect,
        "facetLimit": facetLimit,
        "facetOffset": facetOffset,
        "q": q,
    }

    gbif_kwargs = {key: kwargs[key] for key in kwargs if key not in requests_argset}
    if gbif_kwargs is not None:
        xx = dict(
            zip([re.sub("_", ".", x) for x in gbif_kwargs.keys()], gbif_kwargs.values())
        )
        args.update(xx)
    kwargs = {key: kwargs[key] for key in kwargs if key in requests_argset}
    out = gbif_GET(url, args, **kwargs)
    return out






"""Tests for registry module - organizations"""
import vcr
from pygbif import registry


@vcr.use_cassette("test/vcr_cassettes/test_organizations.yaml")
def test_organizations():
    "registry.organizations - basic test"
    res = registry.organizations()
    assert dict == res.__class__
    assert 2 == len(res)
    assert 100 == len(res["data"])
    assert ["data", "meta"] == sorted(res.keys())


@vcr.use_cassette("test/vcr_cassettes/test_organizations_limit.yaml")
def test_organizations_limit():
    "registry.organizations - limit param "
    res = registry.organizations(limit=5)
    assert dict == res.__class__
    assert 5 == len(res["data"])


@vcr.use_cassette("test/vcr_cassettes/test_organizations_uuid.yaml")
def test_organizations_uuid():
    "registry.organizations - with a uuid"
    res = registry.organizations(uuid="e2e717bf-551a-4917-bdc9-4fa0f342c530")
    assert dict == res.__class__
    assert 2 == len(res)
    assert list == res["data"]["contacts"].__class__
    assert dict == res["data"]["contacts"][0].__class__
    assert list == res["data"]["contacts"][0]["position"].__class__
    assert "numPublishedDatasets" in res["data"].keys()



from pygbif.gbifutils import gbif_baseurl, bool2str, gbif_GET

def name_backbone(
   scientificName=None,
   taxonRank=None,
   usageKey=None,
   kingdom=None,
   phylum=None,
   class_=None,
   order=None,
   superfamily=None,
   family=None,
   subfamily=None,
   tribe=None,
   subtribe=None,
   genus=None,
   subgenus=None,
   species=None,
   taxonID=None,
   taxonConceptID=None,
   scientificNameID=None,
   scientificNameAuthorship=None,
   genericName=None,
   specificEpithet=None,
   infraspecificEpithet=None,
   verbatimTaxonRank= None,
   exclude=None,
   strict=None,
   verbose=None,
   checklistKey=None,    
   **kwargs
):
   """
   Match names to the GBIF backbone taxonomy.

   :param scientificName: [str]
      Full scientific name potentially with authorship. (Required)
   :param taxonRank: [str], optional
      Filter by taxonomic rank. See API reference for available values.
   :param usageKey: [str], optional
      The usage key to look up. When provided, all other fields are ignored.
   :param kingdom: [str], optional
      Kingdom to match.
   :param phylum: [str], optional
      Phylum to match.
   :param class_: [str], optional
      Class to match.
   :param order: [str], optional
      Order to match.
   :param superfamily: [str], optional
      Superfamily to match.
   :param family: [str], optional
      Family to match.
   :param subfamily: [str], optional
      Subfamily to match.
   :param tribe: [str], optional
      Tribe to match.
   :param subtribe: [str], optional
      Subtribe to match.
   :param genus: [str], optional
      Genus to match.
   :param subgenus: [str], optional
      Subgenus to match.
   :param species: [str], optional
      Species to match.
   :param taxonID: [str], optional
      The taxon ID to look up. Matches to a taxonID will take precedence over
      scientificName values supplied. A comparison of the matched scientific and
      taxonID is performed to check for inconsistencies.
   :param taxonConceptID: [str], optional
      The taxonConceptID to match. Matches to a taxonConceptID will take precedence
      over scientificName values supplied. A comparison of the matched scientific and
      taxonConceptID is performed to check for inconsistencies.
   :param scientificNameID: [str], optional
      Matches to a scientificNameID will take precedence over scientificName values
      supplied. A comparison of the matched scientific and scientificNameID is performed
      to check for inconsistencies.
   :param scientificNameAuthorship: [str], optional
      The scientific name authorship to match against.
   :param genericName: [str], optional
      Generic part of the name to match when given as atomised parts instead of the full name.
   :param specificEpithet: [str], optional
      Specific epithet to match.
   :param infraspecificEpithet: [str], optional
      Infraspecific epithet to match.
   :param verbatimTaxonRank: [str], optional
      Filters by free text taxon rank.
   :param exclude: [str], optional
      An array of usage keys to exclude from the match.
   :param strict: [bool], optional
      If set to True, fuzzy matches only the given name, but never a taxon in the upper classification.
   :param verbose: [bool], optional
      If set to True, shows alternative matches which were considered but then rejected.
   :param checklistKey: [str], optional
      The key of a checklist to use. Default is the GBIF Backbone taxonomy.

   ``name_backbone()`` return a dictionary with the following keys: 
   ``['usage', 'classification', 'diagnostics', 'synonym']`` 

   - ``usage``: Returns the matched name and some details such usage key.
   - ``classification``: Returns the upper classification of the matched name.
   - ``diagnostics``: Returns information about the match, such as match type, issues, and confidence.
   - ``synonym``: Indicates if the matched name is a synonym.

   The default is to return the best match for the given name. If there are "multiple equal matches", ``name_backbone()``, 
   will return a note in the diagnostics section:
   ``res["diagnostics"]["note"] = "Multiple equal matches for name"``.

   This note usually happens when a binomial name is provided without authorship. Proving authorship will almost always fix the 
   "multiple equal matches" issue. If ``verbose=True``, the function will return other alternative matches. These are accessible 
   via ``res['diagnostics']['alternatives']``.     

   If your name does not get a match, the GBIF API will return ``[matchType] ='NONE'``. If the species-level
   name is not found, the API will sometimes return ``[matchType] = 'HIGHERRANK'``. With higher rank matches,
   the name is matched to a higher taxonomic rank, such as genus or family. Often supplying authorship will 
   improve matching results.  

   If ``strict=True``, then higher taxon ranks will not be returned when there is a "fuzzy match". 
   Higher rank matches will still be returned if the match is exact.

   To match names against a specific checklist, you can use the `checklistKey` parameter.
   This allows you to specify a checklist from which the name should be matched. If no checklistKey is provided,
   the GBIF Backbone Taxonomy is used by default.

   For more information, see the GBIF API documentation:
   https://techdocs.gbif.org/en/openapi/v1/species#/Searching%20names/matchNames

   Usage::

       from pygbif import species
       species.name_backbone(scientificName="Helianthus annuus", kingdom="Plantae")
       species.name_backbone(scientificName="Poa", taxonRank="GENUS", family="Poaceae")

       # Verbose - gives back alternatives
       species.name_backbone(scientificName="Helianthus annuus", kingdom="Plantae", verbose=True)

       # Strictness
       # If strict=True, then higher taxon ranks will not be returned  when there is a "fuzzy match".
       # Higher rank matches will still be returned if the match is exact.
       species.name_backbone(scientificName="Poa", kingdom="Plantae", verbose=True, strict=False)
       species.name_backbone(scientificName="Helianthus annuus", kingdom="Plantae", verbose=True, strict=True)

       # Multiple equal matches
       species.name_backbone(scientificName="Oenante")
       species.name_backbone(scientificName="Oenante", verbose=True)
       species.name_backbone(scientificName="Calopteryx")
       species.name_backbone(scientificName="Calopteryx", verbose=True)

       # Including authorship in scientificName fixes "Multiple equal matches" note
       species.name_backbone(scientificName="Calopteryx splendens (Harris, 1780)")
       species.name_backbone(scientificName="Oenanthe L.")

       # Match using an alternative checklist 
       species.name_backbone(scientificName="Calopteryx splendens", checklistKey="7ddf754f-d193-4cc9-b351-99906754a03b")
   """
   url = "https://api.gbif.org/v2/" + "species/match"
   args = {
       "scientificName": scientificName,
       "taxonRank": taxonRank,
       "usageKey": usageKey,
       "kingdom": kingdom,
       "phylum": phylum,
       "class": class_,
       "order": order,
       "superfamily": superfamily,
       "family": family,
       "subfamily": subfamily,
       "tribe": tribe,
       "subtribe": subtribe,
       "genus": genus,
       "subgenus": subgenus,
       "species": species,
       "taxonID": taxonID,
       "taxonConceptID": taxonConceptID,
       "scientificNameID": scientificNameID,
       "scientificNameAuthorship": scientificNameAuthorship,
       "genericName": genericName,
       "specificEpithet": specificEpithet,
       "infraspecificEpithet": infraspecificEpithet,
       "verbatimTaxonRank": verbatimTaxonRank,
       "exclude": exclude,
       "strict": bool2str(strict),
       "verbose": bool2str(verbose),
       "checklistKey": checklistKey
   }
   tt = gbif_GET(url, args, **kwargs)
   return tt